# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from pygcam.log import getLogger
from .McsSubcommandABC import McsSubcommandABC

_logger = getLogger(__name__)

# TBD: Database initialization now occurs in getDatabase() if needed. This
# TBD: command is semantically more like "delsim", which both deletes and
# TBD: recreates the database schema and removes the simulation directory

def driver(args, tool):
    import os
    '''
    Set up the SQL database. Must call startDb() before calling this.

    '''
    import shutil

    from pygcam.config import getParam
    from ..Database import getDatabase
    from ..error import PygcamMcsSystemError

    if args.deleteSims:
        # Remove the whole sims dir and remake it
        runSimsDir = getParam('MCS.RunSimsDir')
        if os.path.exists(runSimsDir):
            try:
                shutil.rmtree(runSimsDir)
                os.mkdir(runSimsDir)
            except Exception as e:
                raise PygcamMcsSystemError('Failed to recreate sim dir %s: %s' % (runSimsDir, e))

    db = getDatabase()

    # reinitialize the db
    db.initDb(args=args)

    if not (args.empty):
        from ..XMLResultFile import XMLResultFile
        XMLResultFile.addOutputs()


class DelSimCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Initialize the database for the given user application. 
            This is done automatically by the sub-command "new" and should be used only to 
            recreate the database from scratch.'''}
        super(DelSimCommand, self).__init__('delsim', subparsers, kwargs)

    def addArgs(self, parser):
        parser.add_argument('-r', '--deleteSims', action='store_true', default=False,
                            help='Delete all simulations from the run directory.')

        parser.add_argument('-e', '--empty', action='store_true', default=False,
                            help='''Create the database schema but don't add any data.
                            Useful when restoring from a dumped database.''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



'''
.. Created on: 12/1/16
   Map-generation API

.. Copyright (c) 2016 Richard Plevin and University of California
   See the https://opensource.org/licenses/MIT for license details.
'''

from .query import interpolateYears, limitYears
from .queryFile import RewriteSetParser, RewriteSet
from .utils import coercible

def choropleth(df, imagefile, dataCol=None, categorical=False,
               byAEZ=False, title=None, palette=None, animate=False):
    """
    Low-level routine to draw a choropleth from the DataFrame `df`, assumed
    to have a column named "region" that identifies regions to populate based
    on data in `dataCol`. If `animate` is True, an animated gif is generated
    with a frame for each year column in `df`.

    :param df: (pandas.DataFrame) a "GCAM" dataframe, with various attributes
      describing region, sector, subsector, etc., plus values for a number of
      time-steps (possibly interpolated), and a column indicating units.
    :param imagefile: (str) pathname of the image file to create. Format is
      determined by filename extension.
    :param dataCol: (str) The name of a column holding the data to plot.
    :param categorical: (bool) True if data to plot are categorical rather
       than continuous.
    :param byAEZ: (bool) if True, and `regions` is not None, data are plotted
       by region and AEZ. If True and `regions` is None, data are plotted by
       AEZ only. If False, AEZs are not plotted.
    :param title: (str) A title for the figure. None or "" => no title.
    :param palette: (??) Color palette to use
    :param animate: (bool) If True, generate an animated gif with a frame for
       each year.
    :return: none
    """
    pass

def drawMap(df, imagefile, years=None, interpolate=False, sumByYear=False,
            regions=None, regionMap=None, rewriteSet=None, byAEZ=False,
            title=None, palette=None, animate=False):
    """
    Draw a choropleth map representing the data in "GCAM dataframe" `df`,
    optionally interpolating annual values between time-steps, for the
    given `regions`, and if `byAEZ` is True, by AEZ. If `animate` is True,
    figures are rendered for each year in `years` as an animated GIF,
    otherwise, values for `years` are summed and the result represented
    in the figure. If `years` is None, all years provided in `df` are used.

    :param df: (pandas.DataFrame) a "GCAM" dataframe, with various attributes
      describing region, sector, subsector, etc., plus values for a number of
      time-steps (possibly interpolated), and a column indicating units.
    :param imagefile: (str) pathname of the image file to create. Format is
      determined by filename extension.
    :param dataCol: (str) The name of a column holding the data to plot.
    :param years: (int or iterable of 2 values coercible to int) the range of
       years to include in results. A single int N is treated as [N, N].
    :param interpolate: (bool) if True, linearly-interpolate annual values
       between 5-year time-steps.
    :param sumByYear: (bool) if True, values are summed across years and the
       results are plotted. Implies `interpolate` = True.
    :param regions: (list of str) The regions to draw. If None, all regions
       referenced in `df` are plotted. These may be standard GCAM regions or
       aggregate regions defined by `regionMap`.
    :param regionMap: (dict) a dictionary keyed  by aggregate region names,
       with values being a list of standard GCAM region names.
    :param rewriteSet: (str or pygcam.queryFile.RewriteSet) the name of a
       RewriteSet that is read from the file given by config var
       GCAM.RewriteSetsFile, or a RewriteSet instance.
    :param byAEZ: (bool) if True, and `regions` is not None, data are plotted
       by region and AEZ. If True and `regions` is None, data are plotted by
       AEZ only. If False, AEZs are not plotted.
    :param title: (str) A title for the figure. None or "" => no title.
    :param palette: (??) Color palette to use
    :param animate: (bool) If True, generate an animated gif with a frame for
       each year.
    :return: none
    """
    df = df.copy()  # so we can perform destructive operations on the df

    if rewriteSet:
        regionMap = rewriteSet.asRegionMap() if isinstance(rewriteSet, RewriteSet) \
            else RewriteSetParser.getRegionMap(rewriteSet)

    # Allow 'years' to be an int or pair of ints
    y = coercible(years, int, raiseError=False)
    if y is None:
        startYear, endYear = years
    else:
        startYear = endYear = y

    limitYears(df, (startYear, endYear))

    if interpolate:
        df = interpolateYears(df, startYear=startYear, inplace=True)

    if sumByYear:
        yearCols = filter(str.isdigit, df.columns)
        dataCol = 'sum'
        df[dataCol] = df[yearCols].sum(axis=1)
        animate = False

    if regions:
        # remove regions that are not in the list
        df = df.query('region not in %s' % regions)

    choropleth(df, imagefile, dataCol=dataCol, categorical=False,
               regions=('United States', 'Canada', 'Brazil', 'Rest of Latin America'),
               regionMap=regionMap, byAEZ=byAEZ, title=title,
               palette=palette, animate=animate)



# Created on July 5, 2013
#
# Copyright (c) 2013-2017. The Regents of the University of California (Regents).
# See the file COPYRIGHT.txt for details.

from collections import OrderedDict, defaultdict
from lxml import etree as ET
from math import ceil
import numpy as np
import os
import pandas as pd

from ..config import getParam
from ..log import getLogger
from ..utils import importFromDotSpec
from ..XMLFile import XMLFile

from .Database import getDatabase
from .distro import DistroGen
from .error import PygcamMcsUserError, PygcamMcsSystemError, DistributionSpecError
from .util import mkdirs, loadObjectFromPath, symlink
from .XML import XMLWrapper, findAndSave
from .XMLConfigFile import XMLConfigFile

_logger = getLogger(__name__)

# XML parameter file element tags
INFILE_ELT_NAME      = 'InputFile'
PARAMLIST_ELT_NAME   = 'ParameterList'
PARAM_ELT_NAME       = 'Parameter'
QUERY_ELT_NAME       = 'Query'
CORRELATION_ELT_NAME = 'Correlation'
WITH_ELT_NAME        = 'With'
DISTRO_ELT_NAME      = 'Distribution'
DATAFILE_ELT         = 'DataFile'
PYTHON_FUNC_ELT      = 'PythonFunc'
WRITE_FUNC_ELT       = 'WriteFunc'

# TBD: test these
# These attributes of a <Distribution> element are not parameters to the
# random variable distribution itself, so we exclude these from the
# list we pass when creating the RV.
DISTRO_META_ATTRS     = ['name', 'type', 'apply']
DISTRO_MODIF_ATTRS    = ['lowbound', 'highbound', 'updatezero']


def getBooleanXML(value):
    """
    Get a value from an XML file and convert it into a boolean True or False.

    :param value: any value (it's first converted to a string)
    :return: True of the value is in ['true', '1'], False if the value
             is in ['false', '0']. An exception is raised if any other
             value is passed.
    :raises: PygcamException
    """
    false = ["false", "0"]
    true  = ["true", "1"]

    val = str(value).strip()
    if val not in true + false:
        raise PygcamMcsUserError("Can't convert '%s' to boolean; must be in {false,no,0,true,yes,1} (case sensitive)." % value)

    return (val in true)


class XMLCorrelation(XMLWrapper):
    """
    Store information from a <Correlation> element
    """
    instances = []

    @classmethod
    def createCorrelations(cls, element, param):
        """
        A <Correlation> element can have 1 or more "With" elements specifying
        what to correlate with, e.g., <With name="otherParam">0.6</With>
        """
        targets = element.findall(WITH_ELT_NAME)
        for target in targets:
            obj = XMLCorrelation(element, param, target)
            cls.instances.append(obj)

    @classmethod
    def decache(cls):
        cls.instances = []

    def __init__(self, element, param, target):
        super(XMLCorrelation, self).__init__(element)
        self.param = param
        self.paramName = self.param.getName()
        self.otherName = target.get('name')

        try:
            self.value = float(target.text)

        except ValueError:
            raise PygcamMcsUserError("In parameter %s, the value of a Correlation element (%s) is not numeric"
                                     % (self.paramName, element.text))

        _logger.debug('Defined correlation of %.3f between %s and %s',
                       self.value, self.paramName, self.otherName)

    @classmethod
    def corrMatrix(cls):
        """
        Generate a correlation matrix representing the correlation definitions,
        or None if no Correlation definitions were found.
        """
        if not cls.instances:
            return None

        params = XMLParameter.getInstances()
        count  = len(params)

        corrMat = np.zeros((count, count), dtype=float)

        for i in xrange(count):
            corrMat[i][i] = 1.0  # diagonal must be 1

        for corr in cls.instances:
            p1 = corr.param
            p2 = corr.other
            coef = corr.value

            src1 = p1.getDataSrc()
            src2 = p2.getDataSrc()

            # TBD: Restrict correlation to Distribution, or ok with PythonFunc or DataFile?

            if src1.isShared() != src2.isShared():
                raise PygcamMcsUserError('''
    Parameters '%s' and '%s' cannot be correlated because
    one is 'shared' and the other is 'independent'.''' % (corr.paramName, corr.otherName))

            if src1.isShared():
                varNum1 = p1.rv.getVarNum()
                varNum2 = p2.rv.getVarNum()

                if varNum1 == varNum2:
                    raise PygcamMcsUserError('''
    Parameter '%s' has a single (shared) random variable; it cannot be self-correlated.''' % corr.paramName)
                corrMat[varNum1][varNum2] = corrMat[varNum2][varNum1] = coef

            else:
                # If the parameters have independent RVs per XML element, two parameters
                # can be correlated if they have the same number of RVs, and the RVs of
                # a single param can be correlated pairwise amongst themselves.
                if p1 == p2:
                    raise PygcamMcsUserError('''
    Parameter '%s' has %d independent random variables; it cannot be self-correlated.''' % (corr.paramName, len(p1.getVars())))
                    # Create pairwise correlations among the RVs. This is probably a bad idea if
                    # the query returns more than a few dozen elements, as is usually the case.
                    # vars = p1.getVars()
                    #
                    # from itertools import permutations
                    # # create all unique pairwise combinations of parameters
                    # # and then map these to pairwise combinations of var nums
                    # varNumPairs = map(lambda pair: (pair[0].rv.getVarNum(),
                    #                                 pair[1].rv.getVarNum()),
                    #                   permutations(vars))
                else:
                    # Two different parameters with independent RVs.
                    p1Vars = p1.getVars()
                    p2Vars = p2.getVars()

                    if len(p1Vars) != len(p2Vars):
                        raise PygcamMcsUserError('''
    Parameters '%s' and '%s' cannot be correlated because
    they have a different number of random variables.''' % (corr.paramName, corr.otherName))

                    varNumPairs = map(lambda pair: (pair[0].rv.getVarNum(),
                                                    pair[1].rv.getVarNum()),
                                      zip(p1Vars, p2Vars))

                # In either case, use the pairs of varNums to set up the matrix
                for i, j in varNumPairs:
                    corrMat[i][j] = corrMat[j][i] = coef

        return corrMat

    @classmethod
    def finishSetup(cls):
        for obj in cls.instances:
            obj.other = XMLParameter.getInstance(obj.otherName)
            if obj.other is None:
                raise PygcamMcsUserError('Unknown parameter "%s" in Correlation for %s' % (obj.otherName, obj.paramName))


class XMLQuery(XMLWrapper):
    """Wraps a <Query> element, which holds an xpath expression."""
    def __init__(self, element):
        super(XMLQuery, self).__init__(element)
        self.xpath = element.text.strip()

    def getXPath(self):
        return self.xpath

    def runQuery(self, tree):
        """
        Run an XPath query on the given tree and return a list of the elements found.
        """
        xpath = self.getXPath()
        found = tree.xpath(xpath)
        return found


class XMLTrialData(XMLWrapper):
    """
    An abstract class that provides the protocol required of classes that can produce
    trial data, i.e., XMLDistribution, XMLDataFile, or XMLPythonFunc
    """
    def __init__(self, element, param):
        super(XMLTrialData, self).__init__(element)
        self.param = param
        self.rv    = None

    def getMode(self):
        """
        Get the containing parameter mode, one of {'direct', 'shared', 'independent'}
        """
        return self.param.getMode()

    def isShared(self, mode=None):
        if not mode:
            mode = self.getMode()
        return mode == 'shared'

    def getApply(self):
        """
        Default distro application is 'direct', i.e., directly set the element value.
        """
        s = self.element.get('apply', 'direct')
        return s.lower()

    def isDirect(self, applyAs=None):
        if not applyAs:
            applyAs = self.getApply()
        return applyAs in ('direct', 'replace')     # synonyms, though "direct" might be deprecated

    def isFactor(self, applyAs=None):
        if not applyAs:
            applyAs = self.getApply()
        return applyAs in ('mult', 'multiply')

    def isDelta(self, applyAs=None):
        if not applyAs:
            applyAs = self.getApply()
        return applyAs == 'add'

    def isTrialFunc(self):      # overridden in XMLDistribution when apply is a func reference
        return False

    def isLinked(self):         # overridden in XMLDistribution
        return False

    def ppf(self, *args):
        raise PygcamMcsSystemError('Called abstract "ppf" method of %s' % self.___class__.name)


class XMLDistribution(XMLTrialData):
    """
    A wrapper around a <Distribution> element to provide some
    convenience functions.
    """
    trialFuncDir = None

    @classmethod
    def decache(cls):
        cls.trialFuncDir = None

    def __init__(self, element, param):
        super(XMLDistribution, self).__init__(element, param)

        self.argDict  = {}
        self.modDict = defaultdict(lambda: None)
        self.modDict['apply'] = 'direct'    # set default value

        # Distribution attributes are {apply, highBound, lowBound, updateZero}
        for key, val in element.items():
            key = key.lower()
            self.modDict[key] = val         # TBD: implement these! (currently commented out of parameter-schema.xsd)

        # Load trial function, if defined (i.e., if there's a '.' in the value of the "apply" attribute.)
        self.trialFunc = self.loadTrialFunc()

        self.otherArgs = None
        if self.trialFunc:
            elt = self.element.find('OtherArgs')
            if elt is not None and elt.text:
                codeStr = "dict({})".format(elt.text)
                try:
                    self.otherArgs = eval(codeStr)

                except SyntaxError as e:
                    raise DistributionSpecError("Failed to evaluate expression {}: {}".format(codeStr, e))

        self.child = element[0]

        # TBD: hack alert to deal with non-float attribute. Fix this in the rewrite.
        # TBD: Might be cleaner to have something that isn't confused with a distribution?
        self._isLinked = self.child.tag == 'Linked'
        if self._isLinked:
            attr = 'parameter'
            self.argDict[attr] = self.child.get(attr)
        else:
            for key, val in self.child.items():
                self.argDict[key] = float(val)    # these form the signature that identifies the distroElt function

        self.distroName = self.child.tag.lower()
        sig = DistroGen.signature(self.distroName, self.argDict.keys())
        _logger.debug("<Distribution %s, %s>", ' '.join(map(lambda pair: '%s="%s"' % pair, element.items())), sig)

        gen = DistroGen.generator(sig)
        if gen is None:
            raise DistributionSpecError("Unknown distribution signature %s" % str(sig))

        self.rv = gen.makeRV(self.argDict)  # generate a frozen RV with the specified arguments

    def getKeys(self):
        return self.child.keys()

    def isTrialFunc(self):
        return bool(self.trialFunc)

    def isLinked(self):
        return self._isLinked

    def linkedParameter(self):
        return self.argDict['parameter'] if self._isLinked else None

    def ppf(self, *args):
        """
        ppf() takes a first arg that can be a scalar value (percentile) or a list
        of percentiles, which is how we use it in this application.
        """
        return self.rv.ppf(*args)

    def loadTrialFunc(self):

        funcRef = self.modDict['apply']
        if not '.' in funcRef:
            return None

        if not self.trialFuncDir:
            self.trialFuncDir = getParam('MCS.TrialFuncDir')

        modname, objname = funcRef.rsplit('.', 1)
        modulePath = os.path.join(self.trialFuncDir, modname + '.py')
        try:
            trialFunc = loadObjectFromPath(objname, modulePath)
        except Exception as e:
            raise PygcamMcsUserError("Failed to load trial function '{}': {}".format(funcRef, e))

        return trialFunc

class XMLPythonFunc(XMLTrialData):
    """
    Implements user-defined Python function to be called to generate a series of
    values. The func can access parameters by calling XMLParameter.getInstances(),
    or XMLParameter.getInstance(name) to get one by name.

    The XMLSchema ensures the format is pkg.mod.submod.func or a legal variant.
    """
    def __init__(self, element, param):
        super(XMLPythonFunc, self).__init__(element, param)
        self.func = importFromDotSpec(element.text)

    def ppf(self, *args):
        """
        Call the function specified in the XML file, passing in the arguments to
        the ppf(), which should be a list of percentile values for which to return
        values from whatever underlies this Python function.

        :param args: the list of percentile values for which to return values. These
          can be ignored, e.g., if the function computes some values based on other
          trial data.
        :return: array-like
        """
        return self.func(self.param, *args)


class XMLDataFile(XMLTrialData):
    """
    Holds DataFrames representing files already loaded, keyed by abs pathname.
    A single data file can hold vectors of values for multiple Parameters; this
    way the file is loaded only once. This can be used to load pre-generated
    trial data, e.g. exported from other software.
    """
    cache = {}

    @classmethod
    def getData(cls, pathname):
        """
        Return the DataFrame created by loading the file at pathname. Read the
        file if it's not already in the cache, otherwise, just return the stored DF.
        """
        pathname = os.path.abspath(pathname)    # use canonical name for cache

        if pathname in cls.cache:
            return cls.cache[pathname]

        df = pd.read_table(pathname, sep='\t', index_col=0)
        df.reset_index(inplace=True)
        cls.cache[pathname] = df
        return df

    @classmethod
    def decache(cls):
        cls.cache = {}

    def __init__(self, element, param):
        super(XMLDataFile, self).__init__(element, param)
        filename = self.getFilename()
        self.df  = self.getData(filename)

    def getFilename(self):
        return os.path.expanduser(self.element.text)

    def ppf(self, *args):
        'Pseudo-ppf that just returns a column of data from the cached dataframe.'
        name   = self.param.getName()
        count  = len(args[0])

        if name not in self.df:
            raise PygcamMcsUserError("Variable '%s' was not found in '%s'" % self.getFilename())

        vector = self.df[name]

        if len(vector) < count:
            # get the smallest integer number of repeats of the
            # vector to meet or exceed the desired count
            repeats = int(ceil(float(count)/len(vector)))
            vector  = np.array(list(vector) * repeats)

        return vector[:count]


class XMLVariable(XMLWrapper):
    """
    Simple variable that wraps an Element and provides methods to get, set,
    cache, and restore Element values. This is subclassed by XMLRandomVar
    to provide a ppf() function for generating samples.
    """
    def __init__(self, element, param, varNum=None):
        super(XMLVariable, self).__init__(element)
        self.param       = param
        self.paramPath   = None
        self.storedValue = None      # cached float value

        # N.B. "is not None" is required because "not element" isn't guaranteed to work.
        if element is not None:      # For shared RVs, XMLParameter creates XMLRandomVar with element=None
            self.storeFloatValue()   # We refer to stored value when updating the XML tree for each trial

        # Save the column in the trailMatrix with our values. This is passed
        # in when XMLVariables are instantiated directly (for shared variables)
        # and set automatically during initialization of XMLRandomVariables.
        self.varNum = varNum

    def getParameter(self):
        return self.param

    def getVarNum(self):
        return self.varNum

    def setVarNum(self, varNum):
        self.varNum = varNum

    def getValue(self):
        return self.element.text

    def setValue(self, value):
        """
        Set the value of the element stored in an XMLVariable.
        """
        elt = self.getElement()
        elt.text = str(value)

    def storeFloatValue(self):
        """
        Store the current value as a float
        """
        value = self.getValue()
        try:
            self.storedValue = float(value)
        except Exception:
            name = self.param.getName()
            raise PygcamMcsSystemError("Value '%s' for parameter %s is not a float" % (value, name))

    def getFloatValue(self):
        return self.storedValue


class XMLRandomVar(XMLVariable):
    """
    Stores pointer to an XMLDistribution (unless this is a shared RV).
    Provides access to values by trial number via valueFunc.
    """
    instances = []

    @classmethod
    def saveInstance(cls, obj):
        # Our position in the RV list identifies our column in the trial matrix
        obj.setVarNum(len(cls.instances))
        cls.instances.append(obj)

    @classmethod
    def getInstances(cls):
        return cls.instances

    @classmethod
    def decache(cls):
        cls.instances = []

    def __init__(self, element, param):
        super(XMLRandomVar, self).__init__(element, param)
        self.saveInstance(self)

    def ppf(self, *args):
        """
        Pass-thru to the Distribution's ppf() method. Called by LHS.
        """
        dataSrc = self.param.getDataSrc()
        assert dataSrc, 'Called ppf on shared XMLRandomVar (dataSrc is None)'
        return dataSrc.ppf(*args)

class XMLParameter(XMLWrapper):
    """
    Stores information for a single parameter definition, whether a distribution,
    exogenous data, or a Python function. For <Distribution> and <DataFile>, there
    must be an accompanying <Query> node with an xpath expression.
    """

    # Store a list of all parameter instances for use in generating distributions
    instances = {}

    def __init__(self, element, tree=None):
        super(XMLParameter, self).__init__(element)
        self.tree = tree
        self.active = getBooleanXML(element.get('active', '1'))

        if not self.active:
            return              # nothing else to do

        XMLParameter.saveInstance(self)

        self.vars    = []    # the list of XMLVariable or XMLRandomVar wrapping Elements from query
        self.rv      = None  # stored here only if the distro is shared across Elements from query
        self.query   = None  # XMLQuery instance
        self.dataSrc = None  # A subclass of XMLTrialData instance
        self.parent  = None

        children = filter(lambda elt: not elt.tag is ET.Comment, element.getchildren())
        maxChildren = 3
        assert len(children) <= maxChildren, \
            "<Parameter> cannot have more than %d children. (The XMLSchema is broken.)" % maxChildren

        for elt in children:
            tag = elt.tag

            if tag == QUERY_ELT_NAME:
                self.query = XMLQuery(elt)

            elif tag == CORRELATION_ELT_NAME:
                XMLCorrelation.createCorrelations(elt, self)

            # XMLSchema ensures that the only other tag is Distribution
            # TBD: test for DISTRIBUTION_ELT_NAME explicitly
            else:
                self.child = elt[0]
                childName = self.child.tag

                # Handle the two special cases:
                if childName == PYTHON_FUNC_ELT:
                    cls = XMLPythonFunc
                elif childName == DATAFILE_ELT:
                    cls = XMLDataFile
                else:
                    cls = XMLDistribution   # child provides distro parameters

                self.dataSrc = cls(elt, self)

    @classmethod
    def saveInstance(cls, obj):
        name = obj.getName()
        if name in cls.instances:
            raise PygcamMcsUserError("Error: attempt to redefine parameter '%s'" % name)

        cls.instances[name] = obj

    @classmethod
    def getInstances(cls):
        """
        Return a list of the stored XMLParameter objects
        """
        return cls.instances.values()

    @classmethod
    def decache(cls):
        cls.instances = {}

    @classmethod
    def getParameterLinks(cls):
        linkedParams = filter(lambda param: param.dataSrc.isLinked(), cls.getInstances())
        linkPairs = [(obj.getName(), obj.dataSrc.linkedParameter()) for obj in linkedParams]
        return linkPairs

    @classmethod
    def getInstance(cls, name):
        """
        Find an XMLParameter instance by name
        """
        return cls.instances.get(name, None)

    @classmethod
    def applyTrial(cls, simId, trialNum, df):
        """
        Copy random values for the given trial from the DataFrame df
        into this parameter's elements.
        """
        instances = cls.getInstances()
        for param in instances:
            param.updateElements(simId, trialNum, df)

    def getMode(self):
        s = self.element.get('mode', 'shared')
        return s.lower()

    def isActive(self):
        return self.active

    def hasPythonFunc(self):
        return self.child.tag == PYTHON_FUNC_ELT

    def getDataSrc(self):
        return self.dataSrc

    def getQuery(self):
        return self.query

    # def setTree(self, tree):
    #     self.tree = tree

    def getTree(self):
        return self.tree

    def getVars(self):
        return self.vars

    def setParent(self, parent):
        self.parent = parent

    def runQuery(self, tree):
        """
        Run the XPath query defined for a parameter, save the set of elements found.
        If `tree` is None, the XMLRandomVars are created, but the query isn't run.
        This is used by "gensim" to create the CSV file holding the trial data.
        """
        if not self.query:
            if not self.dataSrc.isTrialFunc():
                raise PygcamMcsUserError("XMLParameter %s has no <Query> and no trial function" % self.getName())

            # trial functions are handled in updateElements()
            self.rv = XMLRandomVar(None, self)
            return

        if tree is None:
            elements = []
        else:
            # Query returns a list of Element instances or None
            elements = self.query.runQuery(tree)
            if elements is None or len(elements) == 0:
                raise PygcamMcsUserError("XPath query '%s' returned nothing for parameter %s" % (self.query.getXPath(), self.getName()))

            _logger.debug("Param %s: %d elements found", self.getName(), len(elements))

        if self.dataSrc.isShared():
            self.rv = XMLRandomVar(None, self)   # Allocate an RV and have all variables share its varNum
            varNum = self.rv.getVarNum()
            vars = map(lambda elt: XMLVariable(elt, self, varNum=varNum), elements)
        else:
            # TBD: this won't work without running the query, but it's not used for now...
            _logger.warn("Called XMLParameter.runQuery without 'tree' for independent element RVs")
            # XMLRandomVar assigns varNum directly
            vars = map(lambda elt: XMLRandomVar(elt, self), elements)

        # Add these to the list since we might be called for multiple scenarios
        self.vars.extend(vars)

    def updateElements(self, simId, trialNum, df):
        """
        Update an element's text (assuming it's a number) by multiplying
        it by a factor, adding a delta, or substituting a given value.
        """
        dataSrc = self.dataSrc
        if dataSrc.isTrialFunc():
            otherArgs = dataSrc.otherArgs or {}
            dataSrc.trialFunc(self, simId, trialNum, df, **otherArgs)
            return

        if not self.vars:
            raise PygcamMcsSystemError("Called updateElements with no variables defined in self.vars")

        isFactor = dataSrc.isFactor()
        isDelta  = dataSrc.isDelta()

        # Update all elements referenced by our list of XMLVariables (or of their subclass, XMLRandomVar)
        for var in self.vars:
            if var.getElement() is None:    # Skip shared RVs, which don't point to an XML element
                _logger.debug("Skipping shared element for %s", var)
                continue

            originalValue = var.getFloatValue()  # apply factor and delta to cached, original value

            varName = var.getParameter().getName()
            randomValue = df.ix[trialNum, varName]

            newValue = randomValue * originalValue if isFactor else \
                ((randomValue + originalValue) if isDelta else randomValue)

            # Set the value in the cached tree so it can be written to trial's local-xml dir
            var.setValue(newValue)


def trialRelativePath(relPath, prefix):
    """
    Convert a pathname that was relative to "exe" to be relative to
    "exe/../../trial-xml" instead.
    For example, "../input/gcam-data-system/aglu-xml/foo.xml"
    becomes "../../trial-xml/input/gcam-data-system/aglu-xml/foo.xml"
    """
    parentDir = '../'
    if not relPath.startswith(parentDir):
        raise PygcamMcsUserError("trialRelativePath: expected path starting with '%s', got '%s'" % \
                                 (parentDir, relPath))

    newPath = os.path.join(prefix, 'trial-xml', relPath[len(parentDir):])
    return newPath


class XMLRelFile(XMLFile):
    """
    A minor extension to XMLFile to store the original relative pathname
    that was indicated in the config file identifying this file.
    """
    def __init__(self, inputFile, relPath, simId):
        from .util import getSimLocalXmlDir

        self.inputFile = inputFile
        self.relPath = relPath

        scenarioDir = getSimLocalXmlDir(simId)
        absPath = os.path.abspath(os.path.join(scenarioDir, relPath))

        super(XMLRelFile, self).__init__(absPath)

    def getRelPath(self):
        return self.relPath

    def getAbsPath(self):
        return self.getFilename()

    def saveSomewhere(self):
        # Save the modified file somewhere for each trial. Maybe in trial-xml?
        pass


class XMLInputFile(XMLWrapper):
    """
    <InputFile> abstraction that represents the XML input file(s) that are associated
    with a given <ScenarioComponent> name. This is necessary because there may be
    different versions of the file for this name in different scenarios. Each instance
    of XMLInputFile holds a set of XMLRelFile instances which each represent an actual
    GCAM input file. Each of these may be referenced from multiple scenarios.
    """

    # Maps a relative filename to an XMLRelFile objects that holds the parsed XML.
    # Exists to ensure that each file is read only once. This is a class variable
    # because the same XML file may be referenced by multiple scenarios and we want
    # to edit each only once.
    xmlFileMap = {}

    @classmethod
    def getModifiedXMLFiles(cls):
        return cls.xmlFileMap.values()

    @classmethod
    def decache(cls):
        cls.xmlFileMap = {}

    def __init__(self, element):
        super(XMLInputFile, self).__init__(element)
        self.parameters = OrderedDict()
        self.element = element
        self.pathMap = defaultdict(set)
        self.xmlFiles = []
        self.writeFuncs = {}    # keyed by funcRef; values are loaded fn objects
        self.writeFuncDir = None

        # User can define functions to call before writing modified XML files.
        # Write function specification is of the form "myPackage.myModule.myFunc".
        # The function must take 1 argument: the XMLInputFile instance. It can make
        # any desired modifications to the XML tree prior to writing it for each trial.
        writeFuncSpec = element.findtext(WRITE_FUNC_ELT, default=[])
        self.loadWriteFunc(writeFuncSpec)

        # Parameters are described once here, but may be applied to multiple files
        self.findAndSaveParams(element)

    def loadWriteFunc(self, funcRef):
        if not funcRef or '.' not in funcRef or funcRef in self.writeFuncs:
            return None

        if not self.writeFuncDir:
            self.writeFuncDir = getParam('MCS.WriteFuncDir')

        modname, objname = funcRef.rsplit('.', 1)
        modulePath = os.path.join(self.writeFuncDir, modname + '.py')
        try:
            fn = loadObjectFromPath(objname, modulePath)
        except Exception as e:
            raise PygcamMcsUserError("Failed to load trial function '%s': %s" % (funcRef, e))

        self.writeFuncs[funcRef] = fn

    def findAndSaveParams(self, element):
       findAndSave(element, PARAM_ELT_NAME, XMLParameter, self.parameters,
                   testFunc=XMLParameter.isActive, parent=self)

    def getComponentName(self):
        """
        Return the name attribute of the element for this <InputFile>, which is a
        unique tag identifying a "logical" file, which may appear in different versions
        in various scenarios, depending on what those scenarios needed to modify.
        """
        name = self.element.get('name')
        return name

    def mergeParameters(self, element):
        """
        Save the parameters associated with the given <InputFile> element within
        self since it holds additional definitions for the same file. This allows
        a single input file to be referenced by multiple groups of parameters.
        """
        eltName = element.get('name')
        compName = self.getComponentName()
        assert compName == eltName, \
            "mergeParameters: InputFile name mismatch: %s and %s" % (compName, eltName)

        self.findAndSaveParams(element)

    def loadFiles(self, context, scenNames, writeConfigFiles=True):
        """
        Find the distinct pathnames associated with our component name. Each scenario
        that refers to this path is stored in a set in self.inputFiles, keyed by pathname.
        The point is to read and update the target file only once, even if referred to
        multiple times.
        """
        import copy

        compName = self.getComponentName()  # an identifier in the config file, e.g., "land2"

        useCopy = not writeConfigFiles  # if we're not writing the configs, use the saved original

        ctx = copy.copy(context)
        simId = context.simId

        for scenName in scenNames:
            ctx.setVars(scenario=scenName)
            configFile = XMLConfigFile.getConfigForScenario(ctx, useCopy=useCopy)

            # If compName ends in '.xml', assume its value is the full relative path, with
            # substitution for {scenario}, e.g., "../local-xml/{scenario}/mcsValues.xml"
            isXML = compName.lower().endswith('.xml')
            relPath = compName if isXML else configFile.getComponentPathname(compName)

            # If another scenario "registered" this XML file, we don't do so again.
            if not relPath in self.xmlFileMap:
                xmlFile = XMLRelFile(self, relPath, simId)
                self.xmlFileMap[relPath] = xmlFile  # unique for all scenarios so we read once
                self.xmlFiles.append(xmlFile)       # per input file in one scenario

            # TBD: In either case, we need to update the config files' XML trees because
            # TBD: some parameter(s) modify the file for this component, in all cases.
            # TBD: This new path has to be coordinated between config file and actual file.
            if writeConfigFiles and not isXML:
                trialRelPath = trialRelativePath(relPath, '../..')
                configFile.updateComponentPathname(compName, trialRelPath)

    def runQueries(self):
        """
        Run the queries for all the parameters in this <InputFile> on each of
        the physical XMLRelFiles associated with this XMLInputFile.
        """
        for param in self.parameters.values():
            if not param.isActive():
                continue

            # Run the query on all the files associated with this abstract InputFile
            for xmlFile in self.xmlFiles:
                tree = xmlFile.getTree()
                param.runQuery(tree)        # accumulates elements from multiple files

    def generateRandomVars(self):
        """
        Called during 'gensim' to generate XMLRandomVars, save variable names to
        the DB, and save the CSV file with all trial data. No need to actually
        run the queries at this point.
        """
        tuples = []

        for param in self.parameters.values():
            if param.isActive():
                pname = param.getName()
                desc  = 'Auto-generated parameter'
                tuples.append((pname, desc))
                param.runQuery(None)    # side-effect is to generate XML RVs

        # If a file (e.g., protected land) has only a writeFunc defined, there will be no parameters
        if tuples:
            #_logger.debug("Saving tuples to db: %s", tuples)
            db = getDatabase()
            db.saveParameterNames(tuples)

    def callFileFunctions(self, xmlFile, trialDir):
        """
        Call any defined per-input-file functions to allow arbitrary
        manipulation of XML after applying all distributions.

        :param xmlFile: (XMLFile or subclass) container for file being operated on
        :param trialDir: (str) this trial's trial-dir
        """
        for fn in self.writeFuncs.values():
            try:
                fn(self, xmlFile, trialDir)     # fn can modify self.tree as needed
            except Exception as e:
                raise PygcamMcsUserError("Call to user WriteFunc '%s' failed: %s" % (fn, e))


class XMLParameterFile(XMLFile):
    """
    Represents the overall parameters.xml file.
    """
    def __init__(self, filename):
        super(XMLParameterFile, self).__init__(filename, schemaPath='mcs/etc/parameter-schema.xsd')

        # XMLInputFiles keyed by scenario component name
        inputFiles = self.inputFiles = OrderedDict()

        # Can't use "findAndSave" here since we need to append if a filename is repeated
        for elt in self.tree.iterfind(INFILE_ELT_NAME):
            compName = elt.get('name')
            if compName in inputFiles:
                inputFileObj = inputFiles[compName]
                inputFileObj.mergeParameters(elt)
            else:
                inputFiles[compName] = XMLInputFile(elt)

        # Match all the Correlation "with" names to actual parameter definitions
        XMLCorrelation.finishSetup()

        _logger.debug("Loaded parameter file: %s", filename)

    def loadInputFiles(self, context, scenNames, writeConfigFiles=True):
        """
        Load the input files, for each scenario in scenNames. Scenarios are
        found in {simDir}/{scenName}.
        """
        for inputFile in self.inputFiles.values():
            inputFile.loadFiles(context, scenNames, writeConfigFiles=writeConfigFiles)

        if writeConfigFiles:
            # Writes all modified configs. Config files' XML trees are updated
            # as InputFile elements are processed.
            XMLConfigFile.writeAll(context)

    def getFilename(self):
        return self.filename

    def runQueries(self):
        for obj in self.inputFiles.values():
            obj.runQueries()

    def generateRandomVars(self):
        for obj in self.inputFiles.values():
            obj.generateRandomVars()

    def writeLocalXmlFiles(self, trialDir):
        """
        Write copies of all modified XML files
        """
        for xmlFile in XMLInputFile.getModifiedXMLFiles():
            exeRelPath = xmlFile.getRelPath()
            absPath = trialRelativePath(exeRelPath, trialDir)

            # Ensure that directories down to basename exist
            dirname = os.path.dirname(absPath)
            mkdirs(dirname)

            # TBD: Might be cleaner to call file func on .xml file rather than on tree
            # Call per-InputFile functions, if defined.
            inputFile = xmlFile.inputFile
            inputFile.callFileFunctions(xmlFile, trialDir)

            if os.path.exists(absPath):
                # remove it to avoid writing through a symlink to the original file
                _logger.debug("Removing %s", absPath)
                os.unlink(absPath)

            _logger.info("XMLParameterFile: writing %s", absPath)
            xmlFile.tree.write(absPath, xml_declaration=True, pretty_print=True)

    def dump(self):
        print("Parameter file: %s" % self.getFilename())
        for obj in self.inputFiles.values():
            obj.dump()

def decache():
    '''
    Clear all instance caches so a new run can begin cleanly
    '''
    XMLConfigFile.decache()
    XMLCorrelation.decache()
    XMLDataFile.decache()
    XMLRandomVar.decache()
    XMLParameter.decache()
    XMLInputFile.decache()
    XMLDistribution.decache()



import os
import shutil
from unittest import TestCase

from pygcam.query import readCsv, readQueryResult
from pygcam.diff import computeDifference
from pygcam.utils import QueryResultsDir, mkdirs

class TestDiffCmd(TestCase):
    def setUp(self):
        self.ws = './data/ws'
        self.baseline = 'base-0'
        self.policy   = 'corn-0'
        self.queryDir = 'queryResults'
        self.years = [2015, 2050]

        self.tmpDir = '/tmp/testDiff'
        self.removeTmpDir()         # start clean
        mkdirs(self.tmpDir)

    def tearDown(self):
        self.removeTmpDir()

    def removeTmpDir(self):
        try:
            shutil.rmtree(self.tmpDir)
        except:
            pass

    def getFilename(self, scenario):
        return os.path.join(self.ws, scenario, QueryResultsDir)

    def readPurposeGrown(self, scenario):
        batchDir = self.getFilename(scenario)
        df = readQueryResult(batchDir, scenario, 'Purpose-grown_biomass_production',
                             years=self.years, interpolate=False)
        return df

    def test_computeDifference(self):
        baseDF = self.readPurposeGrown(self.baseline)
        cornDF = self.readPurposeGrown(self.policy)

        computedDiff = computeDifference(baseDF, cornDF)

        infile = os.path.join(self.ws, self.policy, 'diffs', 'expected-Purpose-grown_biomass_production-corn-0-base-0.csv')
        expectedDiff = readCsv(infile, years=self.years, interpolate=False)

        # check that all values are suitably close to zero
        testDiff = computeDifference(expectedDiff, computedDiff)
        yearCols = filter(str.isdigit, testDiff.columns)
        bools = abs(testDiff[yearCols]) > 1e-8
        self.assertFalse(bools.all().all())






# Copyright (c) 2015 Richard Plevin. See the file COPYRIGHT.txt for details.

from copy import copy
import os
from lxml import etree as ET

from ..log import getLogger
from ..XMLFile import XMLFile
from .error import PygcamMcsUserError

_logger = getLogger(__name__)

# The pre-defined sections of GCAM configuration files
COMPONENTS_GROUP = 'ScenarioComponents'
FILES_GROUP   = 'Files'
STRINGS_GROUP = 'Strings'
INTS_GROUP    = 'Ints'
BOOLS_GROUP   = 'Bools'
DOUBLES_GROUP = 'Doubles'

def getSimScenarioDir(context):
    from .util import getSimLocalXmlDir

    localXmlDir = getSimLocalXmlDir(context.simId)
    scenDir = os.path.join(localXmlDir, context.groupDir, context.scenario)
    return scenDir

def getSimConfigFile(context):
    """
    Returns the path to sim's copy of the config.xml file for the given scenario.
    """
    from .util import ConfigFileName
    scenDir = getSimScenarioDir(context)
    configFile = os.path.join(scenDir, ConfigFileName)
    return configFile

# TBD: create interface to load "most local" version of config file.
class XMLConfigFile(XMLFile):
    '''
    Stores information about a GCAM config.xml file
    '''
    instances = {}  # XMLConfigFile instances keyed by scenario name

    def __init__(self, context, useCopy=False):
        '''
        Read and cache a GCAM configuration file in self.tree.
        '''
        self.context = context = copy(context)

        if useCopy:
            path = self.getBackupPath()
        else:
            path = getSimConfigFile(context)
            self.copyOriginal() # make a copy of the config file for use by runsim

        # Default writePath is where we were read from.
        self.writePath = path

        super(XMLConfigFile, self).__init__(path)

    @classmethod
    def decache(cls):
        cls.instances = {}

    @classmethod
    def getConfigForScenario(cls, context, useCopy=False):
        '''
        Return the path to the run-tree version of the config file
        for the given scenario.
        '''
        key = (context.scenario, useCopy)
        try:
            return cls.instances[key]

        except KeyError:
            obj = XMLConfigFile(context, useCopy=useCopy)
            cls.instances[key] = obj
            return obj

    @classmethod
    def writeAll(cls, context):
        """
        Write all configuration files to disk.
        :return:
        """
        for cfg in cls.instances.values():
            configPath = getSimConfigFile(cfg.context)
            cfg.write(path=configPath)

    def copyOriginal(self):
        '''
        Copy config file to xxx/config-original.xml if config.xml is
        newer, so runsim can use the original to generate XML files.
        '''
        import shutil

        configPath = getSimConfigFile(self.context)
        backupPath = self.getBackupPath()

        # Copy only if the backupPath doesn't exist or is older than configPath
        if not os.path.exists(backupPath) or \
               os.path.getctime(backupPath) < os.path.getctime(configPath):
            _logger.debug('Copying to %s', backupPath)
            shutil.copy2(configPath, backupPath)

    def getBackupPath(self):
        pathname = getSimConfigFile(self.context)
        basename, ext = os.path.splitext(pathname)
        return basename + '-original' + ext       # i.e., [path...]/config-original.xml

    def write(self, path=None):
        """
        Write out the modified configuration file tree to the
        same path that it was read from, or to the given path.

        :param path: (str) pathname to write to
        :return:  none
        """
        if path:
            self.writePath = path       # remember the new path if one is given
        else:
            path = self.writePath

        if os.path.exists(path):        # remove it since it might be a symlink and
            os.unlink(path)             # we don't want to write through to the src

        _logger.debug("XMLConfigFile writing %s", path)
        self.tree.write(path, xml_declaration=True, pretty_print=True)

    def getConfigElement(self, name, group):
        '''
        Find the config file component with the specified tag and return
        the corresponding Element.
        '''
        xpath = '//Configuration/%s/Value[@name="%s"]' % (group, name)
        found = self.tree.xpath(xpath)
        return None if len(found) == 0 else found[0]

    def getConfigValue(self, name, group):
        elt = self.getConfigElement(name, group)
        return None if elt is None else elt.text

    def getConfigGroup(self, group):
        groupElt = self.tree.xpath('//Configuration/%s' % group)
        if groupElt is None:
            raise PygcamMcsUserError('getConfigGroup: group "%s" was not found' % group)

        return groupElt[0]

    def updateConfigElement(self, name, group, newName=None, newValue=None):
        '''
        Update the name attribute, the value (element text), or both for the named config element.
        '''
        if not (newValue or newName):
            raise PygcamMcsUserError('updateConfigElement: must provide newTag or text, or both')

        elt = self.getConfigElement(name, group)
        if elt is None:
            raise PygcamMcsUserError('updateConfigElement: element "%s" was not found in group %s' % (name, group))

        if newName:
            elt.set('Name', newName)

        if newValue:
            elt.text = str(newValue)    # in case it was passed as numeric or bool

    def getComponentPathname(self, name):
        pathname = self.getConfigValue(name, COMPONENTS_GROUP)
        if not pathname:
            raise PygcamMcsUserError('getComponentPathname: a file with tag "%s" was not found' % name)
        return pathname

    def updateComponentPathname(self, name, pathname):
        self.updateConfigElement(name, COMPONENTS_GROUP, newValue=pathname)

    def addConfigElement(self, name, group, value):
        '''
        Append a new element with the given name and value to the given group.
        '''
        groupElt = self.getConfigGroup(group)
        elt = ET.SubElement(groupElt, 'Value', name=name)
        elt.text = value
        return elt

    def deleteConfigElement(self, name, group):
        '''
        Remove the Value element with the given name from the given group
        '''
        groupElt = self.getConfigGroup(group)
        elt = self.getConfigElement(name, group)
        if elt is None:
            raise PygcamMcsUserError('deleteConfigElement: element "%s" was not found in group %s' % (name, group))

        groupElt.remove(elt)



import unittest

from pygcam.XMLFile import McsValues


class TestSymlinks(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_read(self):
        filename = 'data/xml/mcsValues-example.xml'
        mcsValues = McsValues(filename)

        regions = mcsValues.regions()

        self.assertTrue(regions, 'regions() returned an empty list')
        self.assertEqual(len(regions), 2, 'Wrong number of regions; should be 2')

        regionSet = set(regions)
        expectedSet = set(['USA', 'China'])
        self.assertEqual(regionSet, expectedSet, 'Expected %s, got %s' % (expectedSet, regionSet))

        chinaValues = mcsValues.values('China')
        self.assertTrue(chinaValues, "Failed to find values for China")

        expectedSet = set([0.25, 1.25, 1.5])
        chinaSet = set(chinaValues.values())
        self.assertTrue(chinaSet == expectedSet, "Expected values %s, got %s" % (expectedSet, chinaSet))


        expectedSet = set(['biodiesel', 'cellulosic ethanol', 'corn ethanol'])
        chinaSet = set(chinaValues.keys())
        self.assertTrue(chinaSet == expectedSet, "Expected names %s, got %s" % (expectedSet, chinaSet))



#!/usr/bin/env python

'''
.. The "gt" (gcamtool) commandline program

.. Copyright (c) 2016-2017 Richard Plevin and UC Regents
   See the https://opensource.org/licenses/MIT for license details.
'''
from pygcam.tool import main
main()



#!/usr/bin/env python
"""
.. Support for running a sequence of operations for a GCAM project
   that is described in an XML file.

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2015 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from ..subcommand import SubcommandABC


class RunCommand(SubcommandABC):
    def __init__(self, subparsers, name='run', help='Run the steps for a project defined in a project.xml file'):
        kwargs = {'help' : help}
        super(RunCommand, self).__init__(name, subparsers, kwargs)

    def addArgs(self, parser):

        parser.add_argument('-a', '--allGroups', action='store_true',
                            help='''Run all scenarios for all defined groups.''')

        parser.add_argument('-D', '--distribute', action="store_true",
                            help='''Run the given scenarios by queueing them independently. If one of
                            the scenarios is a baseline, it is queued first and the remaining scenarios
                            are queued as dependent on the completion of the baseline job.''')

        parser.add_argument('-f', '--projectFile',
                            help='''The XML file describing the project. If set, command-line
                            argument takes precedence. Otherwise, value is taken from config file
                            variable GCAM.ProjectXmlFile, if defined, otherwise the default
                            is './project.xml'.''')

        parser.add_argument('-g', '--group',
                            help='''The name of the scenario group to process. If not specified,
                            the group with attribute default="1" is processed.''')

        parser.add_argument('-G', '--listGroups', action='store_true',
                            help='''List the scenario groups defined in the project file and exit.''')

        parser.add_argument('-k', '--skipStep', dest='skipSteps', action='append',
                            help='''Steps to skip. These must be names of steps defined in the
                            project.xml file. Multiple steps can be given in a single (comma-delimited)
                            argument, or the -k flag can be repeated to indicate additional steps.
                            By default, all steps are run.''')

        parser.add_argument('-K', '--skipScenario', dest='skipScenarios', action='append',
                            help='''Scenarios to skip. Multiple scenarios can be given in a single
                            (comma-delimited) argument, or the -K flag can be repeated to indicate
                            additional scenarios. By default, all scenarios are run.''')

        parser.add_argument('-l', '--listSteps', action='store_true',
                            help='''List the steps defined for the given project and exit.
                            Dynamic variables (created at run-time) are not displayed.''')

        parser.add_argument('-L', '--listScenarios', action='store_true',
                            help='''List the scenarios defined for the given project and exit.
                            Dynamic variables (created at run-time) are not displayed.''')

        parser.add_argument('-n', '--noRun', action='store_true',
                            help='''Display the commands that would be run, but don't run them.''')

        parser.add_argument('-q', '--noQuit', action='store_true',
                            help='''Don't quit if an error occurs when processing a scenario, just
                            move on to processing the next scenario, if any.''')

        parser.add_argument('-s', '--step', dest='steps', action='append',
                            help='''The steps to run. These must be names of steps defined in the
                            project.xml file. Multiple steps can be given in a single (comma-delimited)
                            argument, or the -s flag can be repeated to indicate additional steps.
                            By default, all steps are run.''')

        parser.add_argument('-S', '--scenario', dest='scenarios', action='append',
                            help='''Which of the scenarios defined for the given project should
                            be run. Multiple scenarios can be given in a single (comma-delimited)
                            argument, or the -S flag can be repeated to indicate additional scenarios.
                            By default, all active scenarios are run.''')

        parser.add_argument('--vars', action='store_true', help='''List variables and their values''')

        parser.add_argument('-x', '--sandboxDir',
                            help='''The directory in which to create the run-time sandbox workspace.
                            Defaults to value of {GCAM.SandboxProjectDir}/{scenarioGroup}.''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        from ..project import projectMain
        projectMain(args, tool)



'''
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
import os
import platform
import re
from pkg_resources import resource_string
from backports import configparser
from .error import ConfigFileError, PygcamException

DEFAULT_SECTION = 'DEFAULT'
USR_CONFIG_FILE = '.pygcam.cfg'
USR_DEFAULTS_FILE = '.pygcam.defaults'

PlatformName = platform.system()

_ConfigParser = None

_ProjectSection = DEFAULT_SECTION

# Support for path translations to access docker-mounted host dirs
_PathMap = None
_PathPattern = None     # compiled regex matching any mapped paths

def savePathMap(mapString):
    """
    Save a list of pathname translations (sorted, descending by length)
    for use with docker, mapping host directories to container-mounted
    directories. The function getParam() performs the translations.

    :param mapString: (str) sequence of newline-limited lines, each
       containing a pair of the form "host-path:container-path".

    :return: nothing
    """
    global _PathMap, _PathPattern

    pairStrings = mapString.split()
    pairs = [s.split(':') for s in pairStrings]

    # strip whitespace
    pairs = [map(str.strip, pair) for pair in pairs]

    # process the longest strings first to avoid overlooking long prefixes
    pairs = sorted(pairs, key = lambda pair: len(pair[0]), reverse=True)
    pattern = '|'.join([pair[0] for pair in pairs])

    _PathPattern = re.compile(pattern)
    _PathMap = dict(pairs)


def _translatePath(value):
    """
    Translate a value if it matches _PathPattern.

    :param value: (str) the config value to translate
    :return: (str) the translated value or original if no key was matched
    """
    matches = re.findall(_PathPattern, value)
    if matches:
        for m in sorted(matches, key=len, reverse=True):
            hostPath = m
            contPath = _PathMap[hostPath]
            # print("re.sub({}, {}, {})".format(hostPath, contPath, value))
            value = re.sub(hostPath, contPath, value)

    return value

def getSection():
    return _ProjectSection

def setSection(section):
    """
    Set the name of the default config file section to read from.

    :param section: (str) a config file section name.
    :return: none
    """
    global _ProjectSection
    _ProjectSection = section

def configLoaded():
    return bool(_ConfigParser)

def getConfig(reload=False, allowMissing=False):
    """
    Return the configuration object. If one has been created already via
    `readConfigFiles`, it is returned; otherwise a new one is created
    and the configuration files are read. Applications generally do not
    need to use this object directly since the single instance is stored
    internally and referenced by the other API functions.

    :param: reload (bool) if True, instantiate a new global ConfigParser.
    :param: allowMissing (bool) if True, a missing config file is not
       treated as an error. This is used only when generating documentation,
       e.g., on readthedocs.org.
    :return: a `ConfigParser` instance.
    """
    if reload:
        global _ConfigParser
        _ConfigParser = None

    return _ConfigParser or readConfigFiles(allowMissing=allowMissing)


def _readConfigResourceFile(filename, package='pygcam', raiseError=True):
    try:
        data = resource_string(package, filename)
    except IOError:
        if raiseError:
            raise
        else:
            return None

    data = unicode(data)
    _ConfigParser.read_string(data, source=filename)
    return data


_usingMCS = None

def mcsSentinelFile():
    home = getHomeDir()
    path = os.path.join(home, '.use_pygcam_mcs')
    return path

def setUsingMCS(value):
    global _usingMCS
    _usingMCS = value

def usingMCS():
    """
    Check if the user environment is configured to use pygcam.mcs, which requires
    that the file ~/.use_pycam_mcs exists. This lets gcamtool know whether to load
    the corresponding built-in sub-commands.

    :return: (bool) True if user environment indicates to use pygcam-mcs.
    """
    global _usingMCS

    if _usingMCS is None:
        path = mcsSentinelFile()
        setUsingMCS(os.path.exists(path))

    return _usingMCS

def getHomeDir():
    if PlatformName == 'Windows':
        # HOME exists on all Unix-like systems; for Windows it's HOMEPATH or HOMESHARE.
        # If set, we use PYGCAM_HOME to identify the folder with the config file;
        # otherwise, we use HOMESHARE if set, or HOMEPATH, in that order.
        env = os.environ
        homedir = env.get('PYGCAM_HOME') or env.get('HOMESHARE') or env.get('HOMEPATH')
        drive, path = os.path.splitdrive(homedir)
        drive = drive or env.get('HOMEDRIVE') or 'C:'
        home = os.path.realpath(drive + path)
        home = home.replace('\\', '/')            # avoids '\' quoting issues
    else:
        home = os.getenv('HOME')

    return home


def userConfigPath():
    path = os.path.join(getHomeDir(), USR_CONFIG_FILE)
    return path

def configDefaultsPath():
    path = os.path.join(getHomeDir(), USR_DEFAULTS_FILE)
    return path

# def removeDefaultsFile():
#     """
#     If the defaults file exists, delete it. This is done by the
#     'init' sub-command to avoid stale defaults files.
#     """
#     path = configDefaultsPath()
#     if os.path.exists(path):
#         os.remove(path)


_instructions = '''#
# This file describes variables defined by the pygcam code. These
# variables allow you to customize many aspects of pygcam, such as
# file locations, and arguments to various commands. See the
# documentation for the config module for detailed explanations.
#
# Use this file to understand the role of each variable and to copy
# and paste variables into your .pygcam.cfg file, as desired.
#
# The default configuration values are provided herein. There is no
# need to specify these variables in your .pygcam.cfg unless you want
# to override the defaults.
#
# To set a project-specific value to override the defaults for one
# project, create a new section by indicating the project name in
# square brackets. For example for the project PROJECT0, you would
# add [PROJECT0]. All settings after this until the next section
# declaration (or end of file) taken as values for "PROJECT0". The
# "gt" script allows you to identify the project you are operating
# on so that the corresponding values are read from the config file.
#
'''

def writeSystemDefaultsFile(systemDefaults):
    '''
    If the system defaults file (~/pygcam.defaults) doesn't exist,
    write the system defaults to the file. Otherwise, just return.
    '''
    path = configDefaultsPath()
    if os.path.exists(path):
        return

    content = _instructions + systemDefaults

    # Add dynamically generated vars (as "raw" values so the obey user's settings of referenced variables)
    content += "\n# For up-to-date documentation of configuration variables, see the\n"
    content += "# listing at the end of https://pygcam.readthedocs.io/en/latest/config.html\n\n"
    content += "# User's home directory\nHome = %s\n\n" % getParam('Home', raw=True)
    content += "# Name of gcam executable relative to 'exe' dir\nGCAM.Executable = %s\n\n" % getParam('GCAM.Executable', raw=True)
    content += "# Location of ModelInterface jar file\nGCAM.MI.JarFile = %s\n\n" % getParam('GCAM.MI.JarFile', raw=True)
    content += "# Whether to use a virtual display when running ModelInterface\nGCAM.MI.UseVirtualBuffer = %s\n\n" % getParam('GCAM.MI.UseVirtualBuffer', raw=True)
    content += "# Editor command to invoke by 'gt config -e'\nGCAM.TextEditor = %s\n\n" % getParam('GCAM.TextEditor', raw=True)

    if PlatformName == 'Windows':   # convert line endings from '\n' to '\r\n' for Windows
        content = content.replace(r'\n', '\r\n')

    # create a file with the system defaults
    with open(path, 'w') as f:
        f.write(content)

def setMacJavaVars():
    """
    Set the environment vars java uses to locate libraries. If var is
    already set, don't override it. Must be called after config info
    has been read since it references GCAM.RefWorkspace.
    """
    if not os.environ.get('JAVA_LIB'):
        refWorkspace = getParam('GCAM.RefWorkspace')
        javaLib = os.path.join(refWorkspace, 'libs/java/lib')
        os.environ['JAVA_LIB'] = javaLib
        setParam('$JAVA_LIB', javaLib)

def readConfigFiles(allowMissing=False):
    """
    Read the pygcam configuration files, starting with ``pygcam/etc/system.cfg``,
    followed by ``pygcam/etc/{platform}.cfg`` if present. If the environment variable
    ``PYGCAM_SITE_CONFIG`` is defined, its value should be a config file, which is
    read next. Finally, the user's config file, ``~/.pygcam.cfg``, is read. Each
    successive file overrides values for any variable defined in an earlier file.

    :return: a populated ConfigParser instance
    """
    global _ConfigParser

    # Strict mode prevents duplicate sections, which we do not restrict
    _ConfigParser = configparser.ConfigParser(comment_prefixes=('#'),
                                              strict=False,
                                              empty_lines_in_values=False)

    # don't force keys to lower-case: variable names are case sensitive
    _ConfigParser.optionxform = lambda option: option

    home = getHomeDir()
    _ConfigParser.set(DEFAULT_SECTION, 'Home', home)
    _ConfigParser.set(DEFAULT_SECTION, 'User', os.getenv('USER', 'unknown'))

    # Create vars from environment variables as '$' + variable name, as in the shell
    for name, value in os.environ.iteritems():
        value = value.replace(r'%', r'%%')
        _ConfigParser.set(DEFAULT_SECTION, '$' + name, value)

    # Initialize config parser with default values
    systemDefaults = _readConfigResourceFile('etc/system.cfg')

    # Read platform-specific defaults, if defined. No error if file is missing.
    _readConfigResourceFile('etc/%s.cfg' % PlatformName, raiseError=False)

    # if user is working with pygcam.mcs, we load additional defaults.
    if usingMCS():
        _readConfigResourceFile('mcs/etc/mcs.cfg', raiseError=True)

    siteConfig = os.getenv('PYGCAM_SITE_CONFIG')
    if siteConfig:
        try:
            with open(siteConfig) as f:
               _ConfigParser.read_file(f)
        except Exception as e:
            print("WARNING: Failed to read site config file: %s" % e)

    # Customizations are stored in ~/.pygcam.cfg
    usrConfigPath = userConfigPath()

    # os.path.exists doesn't always work on Windows, so just try opening it.
    try:
        with open(usrConfigPath) as f:
           _ConfigParser.read_file(f)

    except IOError:
        if not allowMissing:
            if not os.path.lexists(usrConfigPath):
                raise ConfigFileError("Missing configuration file %s" % usrConfigPath)
            else:
                raise ConfigFileError("Can't read configuration file %s" % usrConfigPath)

    try:
        # Write the system defaults to ~/.pygcam.defaults if necessary
        defaultsPath = configDefaultsPath()
        if not os.path.exists(defaultsPath):
            writeSystemDefaultsFile(systemDefaults)
    except Exception as e:
        if not allowMissing:
            raise ConfigFileError("Failed to write %s: %s" % (defaultsPath, e))

    # Dynamically set (if not defined) GCAM.ProjectName in each section, holding the
    # section (i.e., project) name. If user has set this, the value is unchanged.
    projectNameVar = 'GCAM.ProjectName'
    for section in getSections():
        if not (_ConfigParser.has_option(section, projectNameVar) and   # var must exist
                _ConfigParser.get(section, projectNameVar)):            # and not be blank
            _ConfigParser.set(section, projectNameVar, section)

    projectName = getParam('GCAM.DefaultProject', section=DEFAULT_SECTION)
    if projectName:
        setSection(projectName)

    # Set up JAVA environment if needed
    if PlatformName == 'Darwin':
        setMacJavaVars()

    return _ConfigParser

def getSections():
    return _ConfigParser.sections()

def getConfigDict(section=DEFAULT_SECTION, raw=False):
    """
    Return all variables defined in `section` as a dictionary.

    :param section: (str) the name of a section in the config file
    :param raw: (bool) whether to return raw or interpolated values.
    :return: (dict) all variables defined in the section (which includes
       those defined in DEFAULT.)
    """

    # Translation function of identity
    func = _translatePath if _PathMap else lambda x: x

    d = {key : func(value) for key, value in _ConfigParser.items(section, raw=raw)}
    return d

def setParam(name, value, section=None):
    """
    Set a configuration parameter in memory.

    :param name: (str) parameter name
    :param value: (any, coerced to str) parameter value
    :param section: (str) if given, the name of the section in which to set the value.
       If not given, the value is set in the established project section, or DEFAULT
       if no project section has been set.
    :return: value
    """
    section = section or getSection()
    _ConfigParser.set(section, name, value)
    return value

def getParam(name, section=None, raw=False, raiseError=True):
    """
    Get the value of the configuration parameter `name`. Calls
    :py:func:`getConfig` if needed.

    :param name: (str) the name of a configuration parameters. Note
       that variable names are case-insensitive. Note that environment
       variables are available using the '$' prefix as in a shell.
       To access the value of environment variable FOO, use getParam('$FOO').

    :param section: (str) the name of the section to read from, which
      defaults to the value used in the first call to ``getConfig``,
      ``readConfigFiles``, or any of the ``getParam`` variants.
    :return: (str) the value of the variable, or None if the variable
      doesn't exist and raiseError is False.
    :raises NoOptionError: if the variable is not found in the given
      section and raiseError is True
    """
    section = section or getSection()

    if not section:
        raise PygcamException('getParam was called without setting "section"')

    if not _ConfigParser:
        getConfig()

    try:
        value = _ConfigParser.get(section, name, raw=raw)

    except configparser.NoSectionError:
        if raiseError:
            raise PygcamException('getParam: unknown section "%s"' % section)
        else:
            return None

    except configparser.NoOptionError:
        if raiseError:
            raise PygcamException('getParam: unknown variable "%s"' % name)
        else:
            return None

    # perform pathname translation for use of .pygcam.cfg in Docker images
    if _PathMap:
        value = _translatePath(value)

    return value

_True  = ['t', 'y', 'true',  'yes', 'on',  '1']
_False = ['f', 'n', 'false', 'no',  'off', '0']

def stringTrue(value, raiseError=True):
    value = str(value).lower()

    if value in _True:
        return True

    if value in _False:
        return False

    if raiseError:
        msg = 'Unrecognized boolean value: "{}". Must one of {}'.format(value, _True + _False)
        raise ConfigFileError(msg)
    else:
        return None


def getParamAsBoolean(name, section=None):
    """
    Get the value of the configuration parameter `name`, coerced
    into a boolean value, where any (case-insensitive) value in the
    set ``{'true','yes','on','1'}`` are converted to ``True``, and
    any value in the set ``{'false','no','off','0'}`` is converted to
    ``False``. Any other value raises an exception.
    Calls :py:func:`getConfig` if needed.

    :param name: (str) the name of a configuration parameters.
    :param section: (str) the name of the section to read from, which
      defaults to the value used in the first call to ``getConfig``,
      ``readConfigFiles``, or any of the ``getParam`` variants.
    :return: (bool) the value of the variable
    :raises: :py:exc:`pygcam.error.ConfigFileError`
    """
    value = getParam(name, section=section)
    result = stringTrue(value, raiseError=False)

    if result is None:
        msg = 'The value of variable "{}", {}, could not converted to boolean.'.format(name, value)
        raise ConfigFileError(msg)

    return result


def getParamAsInt(name, section=None):
    """
    Get the value of the configuration parameter `name`, coerced
    to an integer. Calls :py:func:`getConfig` if needed.

    :param name: (str) the name of a configuration parameters.
    :param section: (str) the name of the section to read from, which
      defaults to the value used in the first call to ``getConfig``,
      ``readConfigFiles``, or any of the ``getParam`` variants.
    :return: (int) the value of the variable
    """
    value = getParam(name, section=section)
    return int(value)

def getParamAsFloat(name, section=None):
    """
    Get the value of the configuration parameter `name` as a
    float. Calls :py:func:`getConfig` if needed.

    :param name: (str) the name of a configuration parameters.
    :param section: (str) the name of the section to read from, which
      defaults to the value used in the first call to ``getConfig``,
      ``readConfigFiles``, or any of the ``getParam`` variants.
    :return: (float) the value of the variable
    """
    value = getParam(name, section=section)
    return float(value)




#!/usr/bin/env python
"""
.. Support for running a sequence of operations for a GCAM project
   that is described in an XML file.

.. Copyright (c) 2015 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from __future__ import print_function
from copy import copy
import glob
import os
import re
import shlex
import sys

from lxml import etree as ET

from .config import getParam, setParam, getConfigDict
from .constants import LOCAL_XML_NAME, XML_SRC_NAME
from .error import PygcamException, CommandlineError, FileFormatError
from .log import getLogger
from .utils import (flatten, shellCommand, getBooleanXML, unixPath,
                    pathjoin, simpleFormat, QueryResultsDir)
from .temp_file import getTempFile
from .XMLFile import XMLFile
from .xmlSetup import ScenarioSetup

_logger = getLogger(__name__)

DefaultProjectFile = './project.xml'

def minWhitespace(text):
    text = text.strip().replace('\n', ' ')
    text = re.sub('\s\s+', ' ', text)
    return text


def dropArgs(args, shortArg, longArg, takesArgs=True):
    args = copy(args)

    # Delete separated versions, e.g., "-s foo" and "--scenario foo"
    for arg in [shortArg, longArg]:
        while arg in args:
            argIndex = args.index(arg)
            if takesArgs:
                del args[argIndex + 1]
            del args[argIndex]

    if takesArgs:
        # Delete contiguous versions, e.g., "-sfoo" and "--scenario=foo"
        matches = filter(lambda s: s.startswith(shortArg) or s.startswith(longArg + '='), args)
        for arg in matches:
            while arg in args:   # in case an arg is repeated...
                args.remove(arg)

    return args

def decacheVariables():
    SimpleVariable.decache()
    _TmpFileBase.decache()

class _TmpFileBase(object):
    """
    Defines features common to _TmpFile and Queries.
    """
    Instances = {}

    def __init__(self, node):
        self.varName = node.get('varName')
        self.delete  = getBooleanXML(node.get('delete',  '1'))

    # Methods to allow subclasses to use superclass' Instances dict.
    # N.B. can't use cls.Instances or each subclass gets own dict.
    @classmethod
    def setInstance(cls, key, value):
        _TmpFileBase.Instances[key] = value

    @classmethod
    def getInstance(cls, key):
        return cls.Instances.get(key)

    @classmethod
    def decache(cls):
        _TmpFileBase.Instances = {}

    @classmethod
    def writeFiles(cls, argDict):
        """
        Write the files and set the associated variables to the generated filenames.
        """
        for obj in cls.Instances.values():
            path = obj.write(argDict)
            argDict[obj.varName] = path

    def write(self, argDict):
        # subclass responsibility
        raise PygcamException("Called abstract _TmpFileBase.write() method")

class _TmpFile(_TmpFileBase):
    """
    Represents the ``<tmpFile>`` element in the projects.xml file.
    """
    def __init__(self, node):
        """
        defaults is an optional _TmpFile instance from which to
        take default file contents, which are appended to or
        replaced by the list defined here.
        """
        super(_TmpFile, self).__init__(node)

        self.replace = getBooleanXML(node.get('replace', '0'))
        self.eval    = getBooleanXML(node.get('eval',    '1'))    # convert {args} before writing file
        self.dir     = node.get('dir')
        self.path    = None

        name = self.varName
        default = self.getInstance(name)  # save default node of the same name, if any
        self.setInstance(name, self)      # replace default with our own definition

        textNodes = node.findall('text')
        defaults = []

        # Substitute any lines based on matching tags, unless we're replacing
        # all the defaults or there are no defaults to replace.
        if not self.replace and default:
            def getTag(node):
                return node.get('tag')

            tags = map(getTag, filter(getTag, textNodes))

            # Save all the defaults that are not overridden by current definitions
            defaults = filter(lambda node: node.get('tag') not in tags, default.textNodes)

        self.textNodes = defaults + textNodes

    def write(self, argDict):
        # Note: TempFiles are deleted in the main driver (tool.py)
        path = getTempFile(suffix='.project.txt', tmpDir=self.dir, delete=self.delete)
        path = unixPath(path)

        with open(path, 'w') as f:
            text = '\n'.join(map(lambda x: x.text or '', self.textNodes)) + '\n'
            if text and self.eval:
                text = simpleFormat(text, argDict)
            f.write(text)

        self.path = path
        return path

class Queries(_TmpFileBase):
    """
    Represents the ``<queries>`` element in the projects.xml file. We don't process the
    <queries> element here; we just store it so we can write it to a temp file as needed.
    Actual reading/processing of contents is handled in queryFile.py.
    """
    def __init__(self, node):
        super(Queries, self).__init__(node)
        self.tree = ET.ElementTree(node)
        self.setInstance(self.varName, self)      # replace default with our own definition

    def write(self, _argDict):
        # Note: TempFiles are deleted in the main driver (tool.py)
        path = getTempFile(suffix='.queries.xml', delete=self.delete)
        path = unixPath(path)
        self.path = path

        self.tree.write(path, xml_declaration=True, pretty_print=True)
        return path

class Step(object):
    maxStep = 0        # for auto-numbering steps lacking a sequence number

    """
    Represents the ``<step>`` element in the projects.xml file.
    """
    def __init__(self, node):
        self.seq    = int(node.get('seq', 0))
        self.name   = node.get('name')
        self.runFor = node.get('runFor', 'all')
        self.group  = node.get('group', None)
        self.optional = getBooleanXML(node.get('optional', 0))
        self.command = minWhitespace(node.text)

        if not self.command:
            raise FileFormatError("<step name='%s'> is missing command text" % self.name)

        if self.seq:
            Step.maxStep = max(self.seq, self.maxStep)
        else:
            Step.maxStep += 1
            self.seq = Step.maxStep

    def __str__(self):
        return "<Step name='%s' seq='%s' runFor='%s'>%s</Step>" % \
               (self.name, self.seq, self.runFor, self.command)

    def run(self, project, baseline, scenario, argDict, tool, noRun=False):
        runFor = self.runFor
        isBaseline = (baseline == scenario.name)
        isPolicy = not isBaseline

        # See if this step should be run.
        if runFor != 'all' and ((isBaseline and runFor != 'baseline') or (isPolicy and runFor != 'policy')):
            return

        # User can substitute an empty command to delete a default step
        if not self.command:
            return

        try:
            command = simpleFormat(self.command, argDict)    # replace vars in template
        except KeyError as e:
            raise FileFormatError("%s -- No such variable exists in the project XML file" % e)

        _logger.info("[%s, %s, %s] %s", scenario.name, self.seq, self.name, command)

        if not noRun:
            if command[0] == '@':       # run internally in gt
                argList = shlex.split(command[1:])
                argList = flatten(map(lambda s: glob.glob(s) or [s], argList))  # expand shell wildcards
                tool.run(argList=argList)
            else:
                shellCommand(command, shell=True)   # shell=True to expand shell wildcards and so on

class SimpleVariable(object):
    """
    Simple variable of name and value to allow loading config file
    variables alongside variables defined in the project XML file.
    """
    Instances = {}

    def __init__(self, name, value, evaluate=False):
        self.name  = name
        self.value = value or ''
        self.eval  = evaluate
        self.Instances[name] = self

    def getValue(self):
        return self.value

    def setValue(self, value):
        self.value = value

    def evaluate(self, argDict):
        return self.getValue()

    @classmethod
    def decache(cls):
        cls.Instances = {}

    @classmethod
    def instances(cls):
        return cls.Instances.values()

    @classmethod
    def definedVars(cls):
        return cls.Instances.keys()


    @classmethod
    def setFromDefault(cls, varName, defaultName):
        vars = cls.Instances
        if not varName in vars:
            vars[varName] = vars[defaultName]

    @classmethod
    def getDict(cls):
        argDict = {name: var.value for name, var in cls.Instances.items()}
        return argDict


class Variable(SimpleVariable):
    """
    Represents the ``<var>`` element in the projects.xml file.
    """
    def __init__(self, node):
        name     = node.get('name')
        evaluate = getBooleanXML(node.get('eval', 0))

        super(Variable, self).__init__(name, node.text, evaluate=evaluate)

    @classmethod
    def evaluateVars(cls, argDict):
        '''Evaluate vars and store results in argDict'''
        for name, var in cls.Instances.items():
            argDict[name] = var.evaluate(argDict)

    def evaluate(self, argDict, value=None):
        # optional arg lets us call it recursively
        value = value or self.getValue()

        result = simpleFormat(value, argDict) if value and self.eval else value

        # recurse in case there are vars whose values are variable references
        return self.evaluate(argDict, value=result) if re.search('\{[^\}]+\}', result) else result

class Project(XMLFile):
    """
    Represents the ``<project>`` element in the projects.xml file.
    """
    def __init__(self, xmlFile, projectName, groupName=None):

        xmlFile = xmlFile or getParam('GCAM.ProjectXmlFile') or DefaultProjectFile

        self.projectName = projectName or getParam('GCAM.DefaultProject')

        if not self.projectName:
            raise CommandlineError("No project name specified and no default project set")

        super(Project, self).__init__(xmlFile, schemaPath='etc/project-schema.xsd', conditionalXML=True)

        self.scenarioGroupName = groupName

        tree = self.tree
        projectNodes = tree.findall('project[@name="%s"]' % projectName)

        if len(projectNodes) == 0:
            raise FileFormatError("Project '%s' is not defined" % projectName)

        if len(projectNodes) > 1:
            raise FileFormatError("Project '%s' is defined %d times" % (projectName, len(projectNodes)))

        projectNode = projectNodes[0]

        self.subdir = projectNode.get('subdir', '')

        defaultsNode = tree.find('defaults')   # returns 1st match
        hasDefaults = defaultsNode is not None

        # Read referenced scenarios.xml file and add it as a child of projectNode
        # If no 'scenariosFile' element is found, use the value of GCAM.ScenarioSetupFile
        nodes = projectNode.findall('scenariosFile')
        if len(nodes) > 1:
            raise FileFormatError("%s: <project> must define at most one <scenariosFile> element" % xmlFile)
        filename = nodes[0].get('name') if len(nodes) == 1 else getParam('GCAM.ScenarioSetupFile')
        setupFile = pathjoin(os.path.dirname(xmlFile), filename)    # interpret as relative to including file
        self.scenarioSetup = ScenarioSetup.parse(setupFile)

        filename = getParam('GCAM.ScenarioSetupOutputFile')
        if filename:
            _logger.debug('Writing "%s"', filename)
            with open(filename, 'w') as stream:
                self.scenarioSetup.writeXML(stream)

        self.scenarioGroupDict = self.scenarioSetup.groupDict
        self.setGroup(groupName)    # if None, resets scenarioGroupName to default group

        dfltSteps = map(Step, defaultsNode.findall('./steps/step')) if hasDefaults else []
        projSteps = map(Step, projectNode.findall('./steps/step'))
        allSteps  = dfltSteps + projSteps

        # project steps with same name and seq overwrite defaults
        self.stepsDict = stepsDict = {}
        for step in allSteps:
            key = "%s-%d" % (step.name, step.seq)
            stepsDict[key] = step

        self.vars = {}

        if hasDefaults:
            map(Variable, defaultsNode.findall('./vars/var'))

        map(Variable,  projectNode.findall('./vars/var'))

        dfltQueriesNodes = defaultsNode.findall('queries') if hasDefaults else []
        projQueriesNodes = projectNode.findall('queries')
        self.queryFiles  = map(Queries, dfltQueriesNodes + projQueriesNodes)

        dfltTmpFileNodes = defaultsNode.findall('tmpFile') if hasDefaults else []
        projTmpFileNodes = projectNode.findall('tmpFile')
        self.tmpFiles = map(_TmpFile, dfltTmpFileNodes + projTmpFileNodes)

    instance = None

    @classmethod
    def readProjectFile(cls, projectName, groupName=None, projectFile=None):

        # return cached project if already read, otherwise read project.xml
        if not cls.instance or cls.instance.projectName != projectName:
            projectFile = projectFile or getParam('GCAM.ProjectXmlFile', section=projectName)
            cls.instance = Project(projectFile, projectName, groupName)

        return cls.instance

    @classmethod
    def defaultGroupName(cls):
        from pygcam.config import getParam
        projectName = getParam('GCAM.ProjectName')
        obj = cls.readProjectFile(projectName)
        return obj.scenarioSetup.defaultGroup

    def setGroup(self, groupName=None):
        groupDict = self.scenarioGroupDict

        groupName = groupName or self.scenarioSetup.defaultGroup

        if groupName not in groupDict:
            raise FileFormatError("Group '%s' is not defined for project '%s'" % (groupName, self.projectName))

        self.scenarioGroupName = groupName
        self.scenarioGroup = scenarioGroup = groupDict[groupName]
        self.baselineName  = scenarioGroup.baseline
        self.scenarioDict  = scenarioGroup.finalDict

        return self.scenarioGroup

    def maybeListProjectArgs(self, args, knownGroups, knownScenarios, knownStepObjs):
        '''
        If user asked to list scenarios, steps, or variables, do so and quit.
        '''
        self.quit = False

        def showList(strings, header, default=None):
            self.quit = True
            if header:
                print(header)
            for s in strings:
                label = ' (default)' if s == default else ''
                print('  ' + s + label)

        if args.listGroups:
            showList(knownGroups, 'Scenario groups:', default=self.scenarioSetup.defaultGroup)

        if args.listScenarios:
            showList(knownScenarios, 'Scenarios:')

        if args.listSteps:
            self.quit = True
            print('Steps:')
            for step in knownStepObjs:
                label = ' (optional)' if step.optional else ''
                print('  ' + step.name + label)

        if args.vars:
            varList = ["%15s = %s" % (name, value) for name, value in sorted(self.argDict.items())]
            showList(varList, 'Vars:')

        if self.quit:
            sys.exit(0)

    def getKnownSteps(self, asTuple=False):
        '''
        Return a list of known steps in seq order, without duplicates.
        '''
        tmpDict = {}    # used to eliminate duplicates

        def uniqStep(step):
            key = step.name
            if tmpDict.get(key):
                return False # already in dict

            tmpDict[key] = 1
            return key       # first time for this name

        self.sortedSteps = sortedSteps = sorted(self.stepsDict.values(), key=lambda node: node.seq)
        knownStepObjs  = [step for step in sortedSteps if uniqStep(step)]
        knownStepNames = [step.name for step in knownStepObjs]
        return (knownStepNames, knownStepObjs) if asTuple else knownStepNames

    def getKnownScenarios(self):
        '''
        Return a list of known scenarios for the current project and scenarioGroup, alpha sorted
        '''
        # sorting by not(node.isBaseline) results in baseline preceding scenarios
        sortedScenarios = sorted(self.scenarioDict.values(), key=lambda node: not node.isBaseline)
        knownScenarios  = map(lambda node: node.name, sortedScenarios)
        return knownScenarios

    def getKnownGroups(self):
        '''
        Return a list of known scenarioGroups for the current project.
        '''
        sortedGroups = sorted(self.scenarioGroupDict.values(), key=lambda node: node.name)
        knownGroups = map(lambda node: node.name, sortedGroups)
        return knownGroups

    def validateProjectArgs(self, userArgs, knownArgs, argName):
        """
        If the user requested steps or scenarios that are not defined, raise an error.

        :param userArgs: a list of the elements (projects, groups, scenarios, steps)
          passed by the user on the command-line.
        :param knownArgs: a list of known elements of the given type
        :param argName: the tag of the XML element
        :return: nothing
        :raises: CommandlineError if the elements requested by the user are not defined in
          the current (project, scenario) context
        """
        unknownArgs = set(userArgs) - set(knownArgs)
        if unknownArgs:
            s = ' '.join(unknownArgs)
            raise CommandlineError("Requested %s do not exist in project '%s', group '%s': %s" % \
                                  (argName, self.projectName, self.scenarioGroupName, s))

    def sortScenarios(self, scenarioSet):
        """
        If a baseline is in the scenario set, move it to the front and return the new list
        """
        scenarios = list(scenarioSet)

        for scenarioName in scenarios:
            scenario = self.scenarioDict[scenarioName]
            if scenario.isBaseline:
                scenarios.remove(scenarioName)
                scenarios.insert(0, scenarioName)
                break

        return scenarios

    def run(self, scenarios, skipScenarios, steps, skipSteps, args, tool):
        """
        Command templates can include keywords curly braces that are substituted
        to create the command to execute in the shell. Variables are defined in
        the <vars> section of the project XML file.
        """
        projectName = self.projectName
        scenarioGroupName = self.scenarioGroupName
        groupDir = scenarioGroupName if self.scenarioGroup.useGroupDir else ''

        # Push the groupName back into config system so vars can use it
        setParam('GCAM.ScenarioGroup', groupDir, section=projectName)

        # Get the text values for all config variables, allowing variables
        # defined in the project to override them.
        cfgDict = getConfigDict(section=projectName)
        for name, value in cfgDict.items():
            SimpleVariable(name, value)

        self.argDict = argDict = Variable.getDict()

        # Add standard variables for use in step command substitutions
        argDict['project']       = projectName
        argDict['projectSubdir'] = subdir = self.subdir
        argDict['baseline']      = argDict['reference'] = baseline = self.baselineName     # baseline is synonym for reference
        argDict['scenarioGroup'] = scenarioGroupName
        argDict['srcGroupDir']   = srcGroupDir = self.scenarioGroup.srcGroupDir or groupDir
        argDict['projectSrcDir'] = pathjoin('..', XML_SRC_NAME,   srcGroupDir, subdir)
        argDict['projectXmlDir'] = pathjoin('..', LOCAL_XML_NAME, groupDir,    subdir)

        argDict['SEP']  = os.path.sep       # '/' on Unix; '\\' on Windows
        argDict['PSEP'] = os.path.pathsep   # ':' on Unix; ';' on Windows

        knownGroups    = self.getKnownGroups()
        knownScenarios = self.getKnownScenarios()
        knownSteps, knownStepObjs = self.getKnownSteps(asTuple=True)

        self.maybeListProjectArgs(args, knownGroups, knownScenarios, knownStepObjs)

        # explicit statement is the only way to run "optional" steps
        explicitSteps = steps or []

        # Set steps / scenarios to all known values if user doesn't specify any
        steps = set(steps or knownSteps) - set(skipSteps or [])
        scenarios = set(scenarios or knownScenarios) - set(skipScenarios or [])

        # Check that the requested scenarios and steps are defined
        self.validateProjectArgs(scenarios, knownScenarios, 'scenarios')
        self.validateProjectArgs(steps,     knownSteps,     'steps')

        quitProgram = not args.noQuit
        run = not args.noRun

        scenarios = self.sortScenarios(scenarios)
        sandboxDir = args.sandboxDir or argDict['GCAM.SandboxDir']

        argDict['baselineDir'] = pathjoin(sandboxDir, baseline)

        # Delete all variants of scenario specification from shellArgs
        # so we can queue these one at a time.
        shellArgs = dropArgs(tool.shellArgs, '-S', '--scenario')
        shellArgs = dropArgs(shellArgs, '-D', '--distribute', takesArgs=False)
        shellArgs = dropArgs(shellArgs, '-a', '--allGroups', takesArgs=False)

        baselineJobId = None

        for scenarioName in scenarios:
            scenario = self.scenarioDict[scenarioName]

            if not scenario.isActive:
                _logger.debug("Skipping inactive scenario: %s", scenarioName)
                continue

            # Construct gt command that does this scenario's steps
            # setting the -S flag for one scenario at a time.
            if args.distribute:
                newArgs = ['+P', projectName] + shellArgs + ['-S', scenarioName] + ['-g', scenarioGroupName]
                jobId = tool.runBatch2(newArgs, jobName=scenarioName, queueName=args.queueName,
                                       logFile=args.logFile, minutes=args.minutes,
                                       dependsOn=baselineJobId, run=run)
                if scenario.isBaseline:
                    baselineJobId = jobId

                continue

            # These get reset as each scenario is processed
            argDict['scenario']       = scenarioName
            argDict['scenarioSubdir'] = scenario.subdir or scenarioName
            argDict['sandboxDir']     = sandboxDir
            argDict['scenarioDir']    = scenarioDir = pathjoin(sandboxDir, scenarioName)
            argDict['diffsDir']       = pathjoin(scenarioDir, 'diffs')
            argDict['batchDir']       = pathjoin(scenarioDir, QueryResultsDir)
            # set in case it wasn't already
            setParam('GCAM.SandboxDir', sandboxDir, section=projectName)

            # Evaluate dynamic variables and re-generate temporary files, saving paths in
            # variables indicated in <tmpFile> or <queries> elements. This is in the scenario
            # loop so run-time variables are handled correctly, though it does result in the
            # files being written multiple times (though with different values.)
            Variable.evaluateVars(argDict)
            _TmpFileBase.writeFiles(argDict)

            try:
                # Loop over all steps and run those that user has requested
                for step in self.sortedSteps:
                    group = step.group
                    if step.name in steps and (not group or                            # no group specified
                                               group == scenarioGroupName or           # exact match
                                               re.match(group, scenarioGroupName)):    # pattern match
                        # Skip optional steps unless explicitly mentioned
                        if (step.optional and step.name not in explicitSteps):
                            continue

                        argDict['step'] = step.name
                        step.run(self, baseline, scenario, argDict, tool, noRun=args.noRun)
            except PygcamException as e:
                if quitProgram:
                    raise
                _logger.error("Error running step '%s': %s", step.name, e)


    def dump(self, steps, scenarios):
        print("Scenario group:", self.scenarioGroupName)
        print("Requested steps:", steps)
        print("Requested scenarios:", scenarios)
        print("Defined steps:", self.getKnownSteps())
        print("Defined scenarios:", self.getKnownScenarios())
        print('Defined vars:')
        for name, var in Variable.Instances.items():
            print("  %15s : %s" % (name, var.getValue()))
        print('\nTmpFiles:')
        for t in self.tmpFiles:
            print("  " + t.varName)

_project = None


def projectMain(args, tool):

    def listify(items):
        '''Convert a list of comma-delimited strings to a single list of strings'''
        return flatten(map(lambda s: s.split(','), items)) if items else None

    steps     = listify(args.steps)
    skipSteps = listify(args.skipSteps)
    scenarios = listify(args.scenarios)
    skipScens = listify(args.skipScenarios)

    project = Project(args.projectFile, args.projectName, args.group)

    groups = project.getKnownGroups() if args.allGroups else [args.group]

    for group in groups:
        project.setGroup(group)
        project.run(scenarios, skipScens, steps, skipSteps, args, tool)



# @author: Richard Plevin
# @author: Sam Fendell
#
# Copyright (c) 2012-2015. The Regents of the University of California (Regents)
# and Richard Plevin. See the file COPYRIGHT.txt for details.

# TBD: eliminate redundancies with pygcam

from __future__ import with_statement, print_function

import os
from inspect import stack, getargspec

from pygcam.config import getParam, getParamAsInt
from pygcam.log import getLogger
from pygcam.utils import mkdirs, pathjoin

from .constants import COMMENT_CHAR
from .context import getSimDir
from .error import BaseSpecError, PygcamMcsUserError, PygcamMcsSystemError

_logger = getLogger(__name__)

_activeYearStrs = None
_activeYearInts = None

YEAR_COL_PREFIX = 'y'

def writeTrialDataFile(simId, df):
    '''
    Save the trial DataFrame in the file 'trialData.csv' in the simDir.
    '''
    simDir = getSimDir(simId)
    dataFile = os.path.join(simDir, 'trialData.csv')

    # If the file exists, rename it trialData.csv-.
    try:
        os.rename(dataFile, dataFile + '-')
    except:
        pass

    df.to_csv(dataFile, index_label='trialNum')


def readTrialDataFile(simId):
    """
    Load trial data (e.g., saved by writeTrialDataFile) and return a DataFrame
    """
    import pandas as pd

    simDir = getSimDir(simId)

    # If SALib version exists, use it; otherwise use legacy file
    dataFile = os.path.join(simDir, 'data.sa', 'inputs.csv')
    if not os.path.lexists(dataFile):
        dataFile = os.path.join(simDir, 'trialData.csv')

    df = pd.read_table(dataFile, sep=',', index_col='trialNum')
    return df
    # return df.as_matrix()

# Deprecated
# def _jobTmpDir():
#     '''
#     Generate the name of a temporary directory based on the value of GCAM.TempDir
#     and the job ID from the environment.
#     '''
#     tmpDir = getParam('GCAM.TempDir')
#     dirName = "mcs.%s.tmp" % 999999         # TBD: getJobNum()
#     dirPath = pathjoin(tmpDir, dirName)
#     return dirPath

def createOutputDir(outputDir):
    from ..utils import removeFileOrTree
    from ..temp_file import getTempDir

    removeFileOrTree(outputDir, raiseError=False)
    tempOutputDir = getParam('MCS.TempOutputDir')

    if tempOutputDir:
        dirPath = getTempDir(suffix='', tmpDir=tempOutputDir, delete=True)
        # dirPath = _jobTmpDir()
        # removeTreeSafely(dirPath, ignore_errors=True)  # rm any files from prior run in this job
        mkdirs(dirPath)
        _logger.debug("Creating '%s' link to %s" % (outputDir, dirPath))

    else:
        mkdirs(outputDir)

def activeYears(asInt=False):
    '''
    Convert a string identifying active years into a list of ints or strs. Values must
    be comma-separated integers or expressions of the form xxxx-yyyy or xxxx-yyyy:z. The
    two expressions indicate ranges of years, with a default timestep of 5 years. If
    given, the final value after the colon indicates an alternative timestep.
    '''
    global _activeYearInts
    global _activeYearStrs

    # return cached values or compute and cache result
    if not _activeYearStrs:
        import re

        def reducer(lst, item):
            m = re.match(r"^(\d{4})-(\d{4})(?::(\d+))?$", item)
            if m:
                start = int(m.group(1))
                end = int(m.group(2))
                step = int(m.group(3)) if m.lastindex == 3 else 5    # default 5 yr timestep
                rng = range(start, end + 1, step)
                lst.extend(rng)
            elif item.isdigit():
                lst.append(int(item))
            else:
                raise PygcamMcsUserError('Element in list of active years is not an integer: %s' % item)

            return lst

        yearStr = getParam('MCS.Years')
        items = yearStr.split(',')
        years = reduce(reducer, items, [])

        _activeYearInts = map(int, years)
        _activeYearStrs = map(str, years)

    return _activeYearInts if asInt else _activeYearStrs

def stripYearPrefix(s):
    '''
    If s is of the form y + 4 digits, remove the y and convert the rest to integer
    '''
    if len(s) == 5 and s[0] == 'y':
        try:
            return int(s[1:])
        except ValueError:
            pass
    return s

def tail(filename, count):
    '''
    Return the last `count` lines from a text file by running the
    tail command as a subprocess. (Faster than native Python...)

    :param filename: a filename
    :param count: number of lines to read from end of file
    :return: a list of the lines read
    '''
    from subprocess import check_output
    pipe = None

    try:
        cmd = 'tail -n %d %s' % (count, filename)
        lines = check_output(cmd, shell=True)

    except Exception as e:
        msg = "Failed run 'tail' on file '%s': %s" % (filename, e)
        raise PygcamMcsSystemError(msg)
    finally:
        if pipe:
            pipe.close()

    return lines

def computeLogPath(simId, scenario, logDir, trials):
    trialMin = min(trials)
    trialMax = max(trials)
    trialRange = trialMin if trialMin == trialMax else "%d-%d" % (trialMin, trialMax)
    jobName  = "%s-s%d-%s" % (scenario, simId, trialRange)   # for displaying in job queue
    logFile  = "%s-%s.out" % (scenario, trialRange)          # for writing diagnostic output
    logPath  = os.path.join(logDir, logFile)
    return logPath, logFile, jobName

def sign(number):
    return cmp(number, 0)

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def checkSuffix(path, suffix):
    return os.path.splitext(path)[1][1:] == suffix

def fileReader(filename, error=BaseSpecError, fileExtension=None, split=True):
    '''
    Generator for reading a file line by line.

    error is an exception class that subclasses BaseSpecError. lineNum is changed
    automatically as the generator reads the file to ensure that any error thrown
    has the correct line number printed.

    fileExtension is the expected extension of the file. If the
    extension doesn't match, an exception is thrown.

    split is a boolean indicating whether the line should be tokenized or not.

    Returns either a list of tokens in a line (if split is True) or a single string.
    '''
    error.filename = filename.replace('\\', '/')

    if fileExtension:
        if not checkSuffix(filename, fileExtension):
            raise error('Wrong file extension. Should be "%s"' % fileExtension)

    with open(filename, 'r') as f:
        for lineNum, line in enumerate(f):
            error.lineNum = lineNum + 1
            line = line.split(COMMENT_CHAR)[0].strip()  # allow comments
            if len(line) == 0:
                continue
            if split:
                line = line.split()
            yield line

def getOptionalArgs(func):
    '''
    Inspects a function to find the names of optional arguments.
    Used in distrogen to generate distrogens from a single function.

    func is a function object.
    Returns a tuple of strings of the optional argument names.
    '''
    args, _, _, defaults = getargspec(func)
    args = args[-len(defaults):]
    return args

def loadModuleFromPath(modulePath):
    '''
    The load a module from a '.py' or '.pyc' file from a path that ends in the
    module name, i.e., from "foo/bar/Baz.py", the module name is 'Baz'.
    '''
    import sys
    from imp import load_source, load_compiled  # lazy import to speed startup

    # Extract the module name from the module path
    base       = os.path.basename(modulePath)
    moduleName = base.split('.')[0]

    try:
        # see if it's already loaded
        module = sys.modules[moduleName]
        _logger.debug('Module %s found in sys.modules: %s' % (moduleName, module))
        return module

    except KeyError:
        pass

    _logger.info('loading module %s' % base)

    # Load the compiled code if it's a '.pyc', otherwise load the source code
    try:
        module = load_source(moduleName, modulePath)
        return module

    except Exception as e:
        errorString = "Can't load module %s from path %s: %s" % (moduleName, modulePath, e)
        _logger.error(errorString)
        raise PygcamMcsUserError(errorString)


def loadObjectFromPath(objName, modulePath, required=True):
    '''
    Load a module and return a reference to a named object in that module.
    If 'required' and the object is not found, an error is raised, otherwise,
    None is returned if the object is not found.
    '''
    module = loadModuleFromPath(modulePath)
    obj    = getattr(module, objName, None)

    if obj or not required:
        return obj

    raise PygcamMcsUserError("Module '%s' has no object named '%s'" % (modulePath, objName))

#
# File and directory utilities for navigating the run-time structure
#
def rmlink(path):
    if os.path.lexists(path) and os.path.islink(path):
        os.remove(path)

def symlink(src, dst):
    rmlink(dst)
    _logger.debug('ln -s %s %s', src, dst)
    try:
        os.symlink(src, dst)
    except Exception:
        print("Can't symlink %s to %s" % (src, dst))
        raise

def rename(direc, src, dest):
    old = os.path.join(direc, src)
    new = os.path.join(direc, dest)
    os.rename(old, new)

def filecopy(src, dst, removeDst=True):
    'Copy src file to dst, optionally removing dst first to avoid writing through symlinks'
    from shutil import copy2        # equivalent to "cp -p"

    _logger.debug("copyfile(%s,%s,%s)" % (src, dst, removeDst))
    if removeDst and os.path.islink(dst):
        os.remove(dst)

    copy2(src, dst)

def copyfiles(files, dstdir, removeDst=True):
    '''
    :param files: a list of files to copy
    :param dstdir: the directory to copy to
    :param removeDst: if True-like, remove destination file before copying
    :return: nothing
    '''
    mkdirs(dstdir)
    for f in files:
        filecopy(f, dstdir, removeDst=removeDst)

def dirFromNumber(n, prefix="", create=False):
    '''
    Compute a directory name using a 2-level directory structure that
    allows 1000 nodes at each level, accommodating up to 1 million files
    (0 to 999,999) in two levels.
    '''
    from numpy import log10     # lazy import

    maxnodes = getParamAsInt('MCS.MaxSimDirs') or 1000

    # Require a power of 10
    log = log10(maxnodes)
    if log != int(log):
        raise PygcamMcsUserError("MaxSimDirs must be a power of 10 (default value is 1000)")
    log = int(log)

    level1 = n / maxnodes
    level2 = n % maxnodes

    directory = os.path.join(prefix, str(level1).zfill(log), str(level2).zfill(log))
    if create:
        mkdirs(directory)

    return directory

def findParamData(paramList, name):
    '''
    Convenience method mostly used in testing.
    Finds a param from a given list of Parameter instances with given name.
    '''
    return filter(lambda x: x.name == name, paramList)[0].data

TRIAL_STRING_DELIMITER = ','

def parseTrialString(string):
    '''
    Converts a comma-separated list of ranges into a list of numbers.
    Ex. 1,3,4-6,2 becomes [1,3,4,5,6,2]. Duplicates are deleted.
    '''
    rangeStrs = string.split(TRIAL_STRING_DELIMITER)
    res = set()
    for rangeStr in rangeStrs:
        r = [int(x) for x in rangeStr.strip().split('-')]
        if len(r) == 2:
            r = range(r[0], r[1] + 1)
        elif len(r) != 1:
            raise ValueError('Malformed trial string.')
        res = res.union(set(r))
    return list(res)

def createTrialString(lst):
    '''
    Assemble a list of numbers into a compact list using hyphens to identify ranges.
    This reverses the operation of parseTrialString
    '''
    from itertools import groupby   # lazy import
    from operator  import itemgetter

    lst = sorted(set(lst))
    ranges = []
    for _, g in groupby(enumerate(lst), lambda (i, x): i - x):
        group = map(lambda x: str(itemgetter(1)(x)), g)
        if group[0] == group[-1]:
            ranges.append(group[0])
        else:
            ranges.append(group[0] + '-' + group[-1])
    return TRIAL_STRING_DELIMITER.join(ranges)

def chunkify(lst, chunks):
    '''
    Generator to turn a list into the number of lists given by chunks.
    In the case that len(lst) % chunksize != 0, all chunks are made as
    close to the same size as possible.
    '''
    l = len(lst)
    numWithExtraEntry = l % chunks  # this many chunks have one extra entry
    chunkSize = (l - numWithExtraEntry) / chunks + 1
    switch = numWithExtraEntry * chunkSize

    i = 0
    while i < l:
        if i == switch:
            chunkSize -= 1
        yield lst[i:i + chunkSize]
        i += chunkSize

def saveDict(d, filename):
    with open(filename, 'w') as f:
        for key, value in d.iteritems():
            f.write('%s=%s\n' % (key, value))

def fullClassname(obj):
    module = obj.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return obj.__class__.__name__
    return module + '.' + obj.__class__.__name__

def isdebugging():
  for frame in stack():
    if frame[1].endswith("pydevd.py"):
      return True
  return False

# Basenames of common files and dirs
ConfigFileName    = "config.xml"
ScenarioFileName  = "scenarios.xml"
ParameterFileName = "parameters.xml"
ResultFileName    = "results.xml"
QueryDirName      = "queries"

SimAppXmlDirName   = "app-xml"
SimLocalXmlDirName = "local-xml"

def getSimXmlFile(simId, filename):
    """
    Returns the path to a file in the sim's XML dir, e.g., {simDir}/app-xml/foo.xml
    """
    simDir = getSimDir(simId)
    xmlDir = os.path.join(simDir, SimAppXmlDirName)
    path = os.path.join(xmlDir, filename)
    return path

def getSimParameterFile(simId):
    """
    Returns the path to sim's copy of the parameters.xml file.
    """
    return getSimXmlFile(simId, ParameterFileName)

def getSimScenarioFile(simId):
    """
    Returns the path to sim's copy of the scenarios.xml file.
    """
    return getSimXmlFile(simId, ScenarioFileName)

def getSimResultFile(simId):
    """
    Returns the path to sim's copy of the results.xml file.
    """
    return getSimXmlFile(simId, ResultFileName)

def getSimLocalXmlDir(simId):
    """
    Returns the path to sim's local-xml dir.
    """
    simDir = getSimDir(simId)
    path = os.path.join(simDir, SimLocalXmlDirName)
    return path

def getRunQueryDir():
    """
    Returns the path to sim's copy of the scenarios.xml file.
    """
    workspace = getParam('MCS.RunWorkspace')

    path = os.path.join(workspace, QueryDirName)
    return path



from __future__ import print_function
from pygcam.xmlEditor import xmlSel, xmlEdit, extractStubTechnology

cfg = './data/xml/configuration_ref.xml'

def testSel(xpath):
    print('%s: %s' % (xpath, xmlSel(cfg, xpath)))

testSel('//ScenarioComponents/Value[@name="socioeconomics"]')
testSel('//ScenarioComponents/Value[@name="industry"]')
testSel('//ScenarioComponents/Value[@name="foobar"]')

def testEdit(pairs):
    xmlEdit(cfg, pairs)

testEdit([("//Bools/Value[@name='PrintPrices']", 1)])
testEdit([("//Files/Value[@name='dbFileName']/@write-output", 0),
          ("//Ints/Value[@name='stop-period']", -1)])


energy_trans = '/Users/rjp/GCAM/current/input/gcam-data-system/xml/energy-xml/en_transformation.xml'
outfile = '/Users/rjp/tmp/extracted.xml'

extractStubTechnology('USA', energy_trans, outfile,  'refining', 'biomass liquids', 'corn ethanol')



'''
.. Support for 'setup' sub-command, which sets up / customizes GCAM project's XML files.

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
import os

from .config import getParam, getParamAsBoolean
from .constants import LOCAL_XML_NAME, DYN_XML_NAME
from .error import SetupException, ConfigFileError
from .log import getLogger
from .utils import copyFileOrTree, removeFileOrTree, mkdirs, pathjoin, symlinkOrCopyFile, removeTreeSafely, pushd

_logger = getLogger(__name__)

# Files specific to different versions of GCAM. This is explicit rather
# than computed so we can easily detect an unknown version number, and
# so that searches for config parameter names find these occurrences.
_VersionSpecificParameterName = {
    '4.2' : 'GCAM.RequiredFiles.4.2',
    '4.3' : 'GCAM.RequiredFiles.4.3',
    '4.4' : 'GCAM.RequiredFiles.4.4',
    '5.0' : 'GCAM.RequiredFiles.5.0',
}

_FilesToCopy = None

def _getVersionSpecificFiles():
    '''
    Compute the list of version-specific files (including universal files)
    that need to be copied or linked into sandboxes, and cache this since
    it doesn't change in a single run.
    '''
    global _FilesToCopy

    if _FilesToCopy:
        return _FilesToCopy

    version = getParam('GCAM.VersionNumber')
    try:
        paramName = _VersionSpecificParameterName[version]
    except KeyError:
        raise ConfigFileError('GCAM.VersionNumber "{}" is unknown. Fix config file or update pygcam.scenarioSetup.py'.format(version))

    versionFiles = getParam(paramName)
    universalFiles = getParam('GCAM.RequiredFiles.All')
    allFiles = universalFiles + ' ' + versionFiles

    fileList = allFiles.split()

    _logger.debug("Version-specific files for GCAM %s: %s", version, fileList)
    _FilesToCopy = fileList
    return fileList

def _getFilesToCopyAndLink(linkParam):
    files = _getVersionSpecificFiles()

    allFiles = set(files)

    # Subtract from the set of all files the ones to link
    toLink = getParam(linkParam)
    filesToLink = toLink.split()
    filesToLinkSet = set(filesToLink)

    unknownFiles = filesToLinkSet - allFiles
    if unknownFiles:
        _logger.warn('Ignoring unknown files specified in %s: %s' % (linkParam, list(unknownFiles)))
        filesToLinkSet -= unknownFiles

    # Copy everything that is not in the filesToLinkSet
    filesToCopy = list(allFiles - filesToLinkSet)
    return filesToCopy, filesToLink

def _remakeSymLink(source, linkname):
    removeFileOrTree(linkname)
    symlinkOrCopyFile(source, linkname)

def _workspaceLinkOrCopy(src, srcWorkspace, dstWorkspace, copyFiles=False):
    '''
    Create a link (or copy) in the new workspace to the
    equivalent file in the given source workspace.
    '''
    # Set automatically on Windows for users without symlink permission
    copyFiles = copyFiles or getParamAsBoolean('GCAM.CopyAllFiles')
    linkFiles = not copyFiles

    if os.path.isabs(src):
        # if absolute path, append only the basename to the runWorkspace
        srcPath = src
        dstPath = pathjoin(dstWorkspace, os.path.basename(os.path.normpath(src)))
    else:
        # if relative path, append the whole thing to both workspaces
        srcPath = pathjoin(srcWorkspace, src)
        dstPath = pathjoin(dstWorkspace, src)

    # Ensure that parent directory exists
    parent = os.path.dirname(dstPath)
    mkdirs(parent)

    # If dstPath is a link, we always remove it and either recreate
    # the link or copy the files as required. If dstPath isn't a
    # link, we remove it only if we are replacing it with a link.
    if os.path.lexists(dstPath) and (linkFiles or os.path.islink(dstPath)):
        removeFileOrTree(dstPath)

    # We've removed dstPath unless we're avoiding re-copying srcPath
    if not os.path.lexists(dstPath):
        if copyFiles:
            _logger.info('Copying %s to %s' % (srcPath, dstPath))
            copyFileOrTree(srcPath, dstPath)
        else:
            symlinkOrCopyFile(srcPath, dstPath)


def createSandbox(sandbox, srcWorkspace=None, forceCreate=False, mcsMode=None):
    '''
    Set up a run-time sandbox in which to run GCAM. This involves copying
    from or linking to files and directories in `workspace`, which defaults
    to the value of config parameter GCAM.SandboxRefWorkspace.

    :param sandbox: (str) the directory to create
    :param srcWorkspace: (str) the workspace to link to or copy from
    :param forceCreate: (bool) if True, delete and recreate the sandbox
    :param mcsMode: ('gensim', 'trial', or None) perform setup appropriate
       for pygcam-mcs trials.
    :return: none
    '''
    if not srcWorkspace:
        srcWorkspace = getParam('GCAM.RefWorkspace') if mcsMode == 'gensim' \
            else getParam('GCAM.SandboxRefWorkspace')

    # MCS "new" sub-command creates its ref workspace; for non-MCS
    # we do it here, on demand, i.e., if it doesn't exist already.
    if not mcsMode:
        copyWorkspace(srcWorkspace)

    if mcsMode and getParamAsBoolean('GCAM.CopyAllFiles'):
        # Not prohibited; just a disk-hogging, suboptimal choice
        _logger.warn('GCAM.CopyAllFiles = True while running MCS')

    _logger.info("Setting up sandbox '%s'", sandbox)

    if os.path.lexists(sandbox) and os.path.samefile(sandbox, srcWorkspace):
        raise SetupException("The run sandbox is the same as the run workspace; no setup performed")

    if forceCreate:
        # avoid deleting the current directory
        with pushd('..'):
            removeTreeSafely(sandbox, ignore_errors=True)
            mkdirs(sandbox)

    # also makes sandbox and sandbox/exe
    logPath = pathjoin(sandbox, 'exe', 'logs')
    mkdirs(logPath)

    filesToCopy, filesToLink = _getFilesToCopyAndLink('GCAM.SandboxFilesToLink')

    for filename in filesToCopy:
        _workspaceLinkOrCopy(filename, srcWorkspace, sandbox, copyFiles=True)

    for filename in filesToLink:
        _workspaceLinkOrCopy(filename, srcWorkspace, sandbox, copyFiles=False)

    outputDir = pathjoin(sandbox, 'output')

    if mcsMode:
        from mcs.util import createOutputDir
        # link {sandbox}/dyn-xml to ../dyn-xml
        dynXmlDir = pathjoin('..', DYN_XML_NAME)

        dynXmlAbsPath = pathjoin(os.path.dirname(sandbox), DYN_XML_NAME)
        mkdirs(dynXmlAbsPath)
        createOutputDir(outputDir) # deals with link and tmp dir...
    else:
        # link {sandbox}/dyn-xml to {refWorkspace}/dyn-xml
        dynXmlDir = pathjoin(srcWorkspace, DYN_XML_NAME)

        # Create a local output dir
        mkdirs(outputDir)

    dynXmlLink = pathjoin(sandbox, DYN_XML_NAME)
    _remakeSymLink(dynXmlDir, dynXmlLink)

    # static xml files are always linked to reference workspace
    localXmlDir  = pathjoin(srcWorkspace, LOCAL_XML_NAME)
    localXmlLink = pathjoin(sandbox, LOCAL_XML_NAME)
    _remakeSymLink(localXmlDir, localXmlLink)


def copyWorkspace(newWorkspace, refWorkspace=None, forceCreate=False, mcsMode=False):
    '''
    Create a copy of a reference workspace by linking to or copying files from
    `refWorkspace`, which defaults to the value of config parameter
    GCAM.RunWorkspace. The copied workspace is the basis for creating sandboxes
    for a Monte Carlo simulation or a non-MCS project.

    :param newWorkspace: (str) the directory to create
    :param refWorkspace: (str) the workspace to link to or copy from (defaults
       to the value of config parameter GCAM.RefWorkspace)
    :param forceCreate: (bool) if True, delete and recreate the sandbox
    :param mcsMode: (bool) if True, perform setup appropriate for gcammcs trials.
    :return: none
    '''
    version = getParam('GCAM.VersionNumber')
    _logger.info("Setting up GCAM workspace '%s' for GCAM %s", newWorkspace, version)

    refWorkspace = refWorkspace or getParam('GCAM.RefWorkspace')

    if os.path.lexists(newWorkspace) and os.path.samefile(newWorkspace, refWorkspace):
        raise SetupException("run workspace is the same as reference workspace; no setup performed")

    # We write a semaphore file when creating Workspace so we can tell
    # if we failed, in which case we forceCreate on the next attempt.
    # The lock just prevents two users from doing this at the same time,
    # though it's probably overkill.
    # Updated 7/10/17: filelocking failed on Cray (NERSC) so was disabled for now.
    mkdirs(newWorkspace)
    semaphoreFile = pathjoin(newWorkspace, '.creation_semaphore')

    try:
        os.remove(semaphoreFile)
        forceCreate = True          # if we can remove it, last attempt failed
    except OSError as e:
        import errno
        if e.errno != errno.ENOENT: # ENOENT => no such file exists
            raise

    if mcsMode and getParamAsBoolean('GCAM.CopyAllFiles'):
        _logger.warn('GCAM.CopyAllFiles = True while running MCS')

    if forceCreate:
        removeTreeSafely(newWorkspace, ignore_errors=True)

    mkdirs(newWorkspace)
    open(semaphoreFile, 'w').close()    # create empty semaphore file

    # Spell out variable names rather than computing parameter names to
    # facilitate searching source files for parameter uses.
    paramName = 'MCS.WorkspaceFilesToLink' if mcsMode else 'GCAM.WorkspaceFilesToLink'

    filesToCopy, filesToLink = _getFilesToCopyAndLink(paramName)

    for filename in filesToCopy:
        _workspaceLinkOrCopy(filename, refWorkspace, newWorkspace, copyFiles=True)

    for filename in filesToLink:
        _workspaceLinkOrCopy(filename, refWorkspace, newWorkspace, copyFiles=False)

    for filename in ['local-xml', 'dyn-xml']:
        dirname = pathjoin(newWorkspace, filename)
        mkdirs(dirname)

    # if successful, remove semaphore
    os.remove(semaphoreFile)



# Copyright (c) 2016 Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from pygcam.log import getLogger
from .McsSubcommandABC import McsSubcommandABC

from ..error import DistributionSpecError

DISCRETE_SUFFIX = 'ddist'
DEFAULT_DISCRETE_PRECISION = 100
DEFAULT_DISCRETE_TOLERANCE = 0.01

SPEC_SEPARATOR  = r'\s+'

_logger = getLogger(__name__)

def getDiscreteDistFromData(data, bins=30):
    """
    Returns a set of (value, frequency) pairs for some title from the given dictionary of {value: count}.
    The values are stored in returnValue[data], and each frequency can be found from minData, minData+binsize,
    minData+2*binsize, etc...
    Assumes there are relatively few discrete values the data can take, and returns
    pairs ready for processing by DiscreteDist.py

    data: dictionary mapping values to counts of those values.

    bins: number of bins to put data into. Default is 30.
    """
    import numpy as np
    from math import ceil, floor
    from ..error import DistributionSpecError

    if bins <= 0 or int(bins) != bins:
        raise DistributionSpecError('Number of bins must be a positive integer.')

    dataArray = np.zeros(bins)
    minData = min(data)
    dataRange = max(data) - minData
    binSize = int(ceil(dataRange / bins))

    # if dataRange is zero, only one value was given. Turns out this edge case isn't well handled by the rest of the code
    if not dataRange:
        return {'data': np.array([1.0]), 'min': minData, 'binSize': 0}

    # In the rare edge case where the range of the data is divisible by the bin number
    # increment binsize to avoid fencemaking error
    if binSize * bins == dataRange:
        binSize += 1
    size = 0
    for key, cnt in data.iteritems():
        dataArray[int(floor((key - minData) / binSize))] += cnt  # divide by size so results sum to 1
        size += cnt

    if size:
        dataArray /= float(size)

    # half bin offset of min because we're sampling from the middle of each bucket, not from its bottom
    return {'data': dataArray, 'min': minData + float(binSize) / 2, 'binSize': binSize}

def getDataDict(fileName, dataTitle, varTitles=None, countTitle=None):
    """
    Takes in a fileName (csv format), rowTitle, and columnTitle and generates
    dataDict with a list of values of data for every value of [varTitle[0], varTitle[1], etc..].
    In other words, each member of varTitle represents a column to key the output dataDict by.
    Returns a dict of dicts. The outer dict is keyed by a tuple constructed from the indices of
    the columns specified by varTitles. The inner dict is keyed by the value for the data in the
    column identified by dataTitle.

    fileName: csv file.

    dataTitle: title of column in file that represents data

    varTitles: List of titles of columns in csv file. Each combination of different values of
    varTitles will be separated into their own dict, and put into the returned dict keyed by a
    tuple made up of values of varTitles.

    countTitle: optional argument. Must be the title of a column of the file. Each entry in
    dataTitle is counted as if it is repeated countTitle times.

    Example: csv file that looks like:

    A,B,C,data,count
    1,2,3,8,10
    2,3,4,7,1
    2,3,4,6,2
    1,2,8,1,3
    2,3,9,7,8

    let varTitles = ['A','B'], dataTitle='data',countTitle='count'
    The returned dataDict will have keys (1,2) and (2,3)
    dataDict[(1,2)] = {8:10, 1:3}
    dataDict[(2,3)] = {7:9, 6:2}
    """
    import csv
    from collections import defaultdict
    from warnings import warn

    if not varTitles:
        warn("dataDict was not given a list of variable names.")
        varTitles = []

    with open(fileName, 'rb') as f:
        reader = csv.reader(f)
        firstLine = reader.next()

        # set appropriate indices
        datIndex = firstLine.index(dataTitle)
        varIndices = [firstLine.index(title) for title in varTitles]

        cntIndex = None
        if countTitle:
            cntIndex = firstLine.index(countTitle)

        # create dataDict
        dataDict = defaultdict(lambda: defaultdict(int))

        for entry in reader:
            dictInd = tuple([entry[i] for i in varIndices])
            cnt = int(entry[cntIndex]) if cntIndex else 1
            # Skip counts of zero, which can occur with the non-forest records in World.csv
            if cnt:
                dataDict[dictInd][float(entry[datIndex])] += cnt

    return dataDict

def driver(args, tool):
    """
    Converts an input csv file into an output file of discrete distribution declarations.

    inputFile: csv file. Must have columns headed with dataTitle, as well as each entry of varTitles and countTitle.

    outputFile: place to store discrete distribution data. Should have suffix matching Distro.DISCRETE_SUFFIX (initially 'out')

    dataTile: Title of column of actual data to be put into discrete distributions.

    varName: Optional name of each entry. If not supplied, varName will be the base name of the file being read from.

    varTitles: Optional array of titles of columns in inputFile. A separate distribution will be made for every value of each column in varTitle.
    For example, if varTitles is ['a','b','c'], and the 'a','b',and 'c' columns of inputFile have domains [1,2],[3,4],[5,6], respectively,
    the outputFile will have 8 rows, one for each of [1,3,5],[1,3,6],[1,4,5],[1,4,6],[2,3,5],[2,3,6],[2,4,5],[2,4,6].

    bins: Number of bins to divide distributions into. Default is 30.
    truncate: Number of decimal places for frequency entries of output file. Default is 3.
    countTitle: Optional column of inputFile. Each entry in dataTitle is treated as if it has been duplicated countTitle number of times.
    keyFunc: Optional function for how to key the entries of the output file. Must be a function of varTitles. Takes in a list
    of the same length as varTitles and outputs a list of the same length.
    Output is a .ddist file that specifies all of the discrete distributions found in the .csv file.
    """
    from os.path import basename

    from ..error import DistributionSpecError
    from ..util import checkSuffix

    inputFile  = args.inputFile
    outputFile = args.outputFile
    dataTitle  = args.dataTitle
    varName    = args.varName
    varTitles  = args.varTitles
    countTitle = args.countTitle
    bins       = args.bins
    truncate   = args.truncate

    if not checkSuffix(outputFile, DISCRETE_SUFFIX):
        raise DistributionSpecError('Output file is not a .%s file' % DISCRETE_SUFFIX)

    if varName is None:
        varName = basename(inputFile).split('.')[0]

    # Get the data from the input file
    dataDict = getDataDict(inputFile, dataTitle, varTitles=varTitles, countTitle=countTitle)

    # Aggregate the data and counts
    distDicts = {key: getDiscreteDistFromData(value, bins) for key, value in dataDict.items()}

    with open(outputFile, 'w') as outFile:
        for key, distDict in distDicts.items():
            outFile.write(varName)
            # Get rid of quotes and whitespace so as not to mess up the reading of .ddist files
            outKey = '[' + ','.join(key) + ']'
            if SPEC_SEPARATOR in outKey:
                raise DistributionSpecError('Key %s contains illegal character "%s".' % (outKey, SPEC_SEPARATOR))
            outFile.write(outKey)
            minValue = distDict['min']
            binSize = distDict['binSize']
            for binNum, frequency in enumerate(distDict['data']):
                f = round(frequency, truncate)
                # If f is too small (rounds to 0), may as well not even print it
                if f:
                    outFile.write('\t' + str(minValue + binNum * binSize) + ':' + str(f))
            outFile.write('\n')

class DiscreteCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Convert csv files to the .ddist format.'''}
        super(DiscreteCommand, self).__init__('discrete', subparsers, kwargs)
        self.group = None   # don't display this for now

    def addArgs(self, parser):
        from ..constants import DEFAULT_BINS, DEFAULT_TRUNCATE, COUNT_TITLE

        parser.add_argument('-i', '--inputFile', required=True,
                            help='Path to input .csv file being converted.')

        parser.add_argument('-o', '--outputFile', required=True,
                            help='Path to output .ddist file.')

        parser.add_argument('-d', '--dataTitle', required=True,
                            help='Actual data title in the .csv file.')

        parser.add_argument('-b', '--bins', type=int, default=DEFAULT_BINS,
                            help='Number of bins to separate discrete distro into')

        parser.add_argument('-t', '--truncate', type=int, default=DEFAULT_TRUNCATE,
                            help='Number of digits to truncate output to. Default is 3')

        parser.add_argument('-c', '--countTitle', type=str, default=COUNT_TITLE,
                            help='Title of column representing counts of data.')

        parser.add_argument('-n', '--varName', type=str, default=None,
                            help='Title of rows of output distribution')

        parser.add_argument('-v', '--varTitles', type=str, nargs='*',
                            help='Titles of columns keying different distributions in the input file')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)


# TODO: maybe simplify using scipy.stats.rv_discrete, but loses the tolerance option.
# http://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_discrete.html
# from scipy import stats
# xk = np.arange(7)
# pk = (0.1, 0.2, 0.3, 0.1, 0.1, 0.0, 0.2)
# custm = stats.rv_discrete(name='custm', values=(xk, pk))

# Deprecated?
class DiscreteDist(object):
    """
    An object to represent a discrete distribution of values.
    """
    def __init__(self, probList, precision=DEFAULT_DISCRETE_PRECISION, tolerance=DEFAULT_DISCRETE_TOLERANCE):
        """
        Takes in a list of (value, probability) tuples to initiate.
        Tolerance allows for distributions with a sum of probabilities close but
        not equal to 1 due to rounding errors. precision determines how many entries
        are in the fast ppf list; ppf values will be accurate within 1/precision
        """
        import numpy as np

        totalProb = sum([x[1] for x in probList])
        if not (1 - tolerance <= totalProb <= 1 + tolerance):
            raise DistributionSpecError('Sum of probabilities != 1 (sum=%f). Try setting the tolerance higher.' % totalProb)

        self.probList = sorted(probList)
        self.values = [x[0] for x in self.probList]
        self.probs = [x[1] / totalProb for x in self.probList]

        # Initialize the lookup table for fast ppf
        self.precision = precision
        self.fastPPF = np.zeros(precision)

        index = 0
        cumProb = 0
        for i in range(precision):
            if i / float(precision) > self.probs[index] + cumProb:
                cumProb += self.probs[index]
                index += 1
            self.fastPPF[i] = self.values[index]

    def __eq__(self, comp):
        return self.values == comp.values and self.probs == comp.probs and self.precision == comp.precision

    def ppf(self, percentiles):
        """
        This ppf function behaves differently than those for continuous
        distributions. In this discrete case, we just want to split the
        range from 0 to 1 so that each discrete value has the assigned
        probability. This way when the LHS function passes in a vector of
        N percentiles, the values returned will occur in the correct proportions.
        The percentiles parameter must be 'array-like' to match (some) of the
        behavior of scipy.stats.rv_continuous.ppf
        """
        if not reduce(lambda x, y: x and 0 < y < 1, percentiles):
            raise DistributionSpecError('Percentiles must all be > 0 and < 1')

        return map(lambda x: self.fastPPF[int(x * self.precision)], percentiles)

    def rvs(self, n=1):
        """Returns one or more random values, according to the RV's distribution."""
        from scipy.stats import uniform

        return self.ppf(uniform.rvs(size=n))
